# coding: utf-8
from atropos.adapters import *
from atropos.align import MatchInfo
from atropos.commands.trim.modifiers import *
from atropos.io.seqio import Sequence
from atropos.util import reverse_complement as rc

DUMMY_ADAPTER = Adapter("ACGT", FRONT)
def front_match(read):
    match = Match(0, 2, 0, 2, 1, 0, True, DUMMY_ADAPTER, read)
    match_info = MatchInfo("read", 0, 0, 2, "", "AC", "ACGTAC", "adapter", "", "##", "######", True, 2, 2, 2)
    return (match, [match_info])
def back_match(read):
    match = Match(6, 8, 6, 8, 1, 0, False, DUMMY_ADAPTER, read)
    match_info = MatchInfo("read", 0, 6, 8, "ACGTAC", "GT", "", "adapter", "######", "##", "", False, 2, 2, 2)
    return (match, [match_info])
    
def test_unconditional_cutter():
    uc = UnconditionalCutter(lengths=[5])
    s = Sequence("read1", 'abcdefg')
    assert UnconditionalCutter(lengths=[2])(s).sequence == 'cdefg'
    assert UnconditionalCutter(lengths=[-2])(s).sequence == 'abcde'
    assert UnconditionalCutter(lengths=[100])(s).sequence == ''
    assert UnconditionalCutter(lengths=[-100])(s).sequence == ''


def test_nend_trimmer():
    trimmer = NEndTrimmer()
    seqs = ['NNNNAAACCTTGGNNN', 'NNNNAAACNNNCTTGGNNN', 'NNNNNN']
    trims = ['AAACCTTGG', 'AAACNNNCTTGG', '']
    for seq, trimmed in zip(seqs, trims):
        _seq = Sequence('read1', seq, qualities='#'*len(seq))
        _trimmed = Sequence('read1', trimmed, qualities='#'*len(trimmed))
        assert trimmer(_seq) == _trimmed


def test_quality_trimmer():
    read = Sequence('read1', 'ACGTTTACGTA', '##456789###')

    qt = QualityTrimmer(10, 10, 33)
    assert qt(read) == Sequence('read1', 'GTTTAC', '456789')

    qt = QualityTrimmer(0, 10, 33)
    assert qt(read) == Sequence('read1', 'ACGTTTAC', '##456789')

    qt = QualityTrimmer(10, 0, 33)
    assert qt(read) == Sequence('read1', 'GTTTACGTA', '456789###')

def test_Modifiers_single():
    m = SingleEndModifiers()
    m.add_modifier(UnconditionalCutter, lengths=[5])
    mod1 = m.get_modifiers(read=1)
    mod2 = m.get_modifiers(read=2)
    assert len(mod1) == 1
    assert isinstance(mod1[0], UnconditionalCutter)
    assert len(mod2) == 0
    # test single-end
    read = Sequence('read1', 'ACGTTTACGTA', '##456789###')
    mod_read = m.modify(read)
    assert len(mod_read) == 1
    assert mod_read[0].sequence == 'TACGTA'

def test_Modifiers_paired_legacy():
    m = PairedEndModifiers(paired="first")
    m.add_modifier(UnconditionalCutter, lengths=[5])
    read1 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
    read2 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
    mod_read1, mod_read2 = m.modify(read1, read2)
    assert mod_read1.sequence == 'TACGTA'
    assert mod_read2.sequence == 'ACGTTTACGTA'

def test_Modifiers_paired_both():
    m = PairedEndModifiers(paired="both")
    m.add_modifier(UnconditionalCutter, read=1|2, lengths=[5])
    mod1 = m.get_modifiers(read=1)
    mod2 = m.get_modifiers(read=2)
    assert len(mod1) == 1
    assert len(mod2) == 1
    assert isinstance(mod1[0], UnconditionalCutter)
    assert isinstance(mod2[0], UnconditionalCutter)
    read1 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
    read2 = Sequence('read1', 'ACGTTTACGTA', '##456789###')
    mod_read1, mod_read2 = m.modify(read1, read2)
    assert mod_read1.sequence == 'TACGTA'
    assert mod_read2.sequence == 'TACGTA'

def test_min_cutter_T_T():
    unconditional_before = UnconditionalCutter((2,-2))
    unconditional_after = UnconditionalCutter((1,-1))
    min_trimmer = MinCutter((5,-5), True, True)
    
    read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
    assert read1.clipped == [0,0,0,0], str(read1.clipped)
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2,2,0,0], str(read1.clipped)
    
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
    
    # test with adapter trimming
    read2 = read1[:]
    read2.sequence = "ATCGAACGTACCG"
    read2.match, read2.match_info = front_match(read2)
    read3 = min_trimmer(read2)
    assert read3.sequence == "TCGAACGTACCG", read3.sequence
    assert read3.clipped == [2,2,1,0]
    
    # test with subsequent clipping
    read4 = unconditional_after(read2)
    assert read4.sequence == "TCGAACGTACC", read4.sequence
    assert read4.clipped == [2,2,1,1], read4.clipped
    read5 = min_trimmer(read4)
    assert read5.sequence == "TCGAACGTACC", read5.sequence
    assert read5.clipped == [2,2,1,1], read5.clipped

def test_min_cutter_F_T():
    unconditional_before = UnconditionalCutter((2,-2))
    unconditional_after = UnconditionalCutter((1,-1))
    min_trimmer = MinCutter((5,-5), False, True)
    
    read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2,2,0,0]
    
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "ATCGATCGAACGTACCG"
    
    # test with adapter trimming
    read2 = read1[:]
    read2.match, read2.match_info = front_match(read2)
    read2.sequence = "CGATCGAACGTACCG"
    read3 = min_trimmer(read2)
    assert read3.sequence == "GAACGTACCG", read3.sequence
    assert read3.clipped == [2,2,5,0]
    
    # test with subsequent clipping
    read4 = unconditional_after(read2)
    assert read4.sequence == "GATCGAACGTACC"
    assert read4.clipped == [2,2,1,1]
    read5 = min_trimmer(read4)
    assert read5.sequence == "GAACGTACC", read5.sequence
    assert read5.clipped == [2,2,5,1]

def test_min_cutter_T_F():
    unconditional_before = UnconditionalCutter((2,-2))
    min_trimmer = MinCutter((4,-4), True, False)
    
    read1 = Sequence('read1', "CAATCGATCGAACGTACCGAT")
    read1 = unconditional_before(read1)
    assert read1.sequence == "ATCGATCGAACGTACCG"
    assert read1.clipped == [2,2,0,0]
    
    # test without adapter trimming
    assert min_trimmer(read1).sequence == "CGATCGAACGTAC"
    
def test_non_directional_bisulfite_trimmer():
    trimmer = NonDirectionalBisulfiteTrimmer(rrbs=True)
    read1 = Sequence('read1', "CAATCGATCGA")
    read2 = Sequence('read2', "CTATCGATC")
    read2.match, read2.match_info = back_match(read2)
    read3 = Sequence('read3', "CTATCGATCCA")
    #assert trimmer(read1).sequence == "ATCGATC"
    assert trimmer(read2).sequence == "CTATCGA"
    assert trimmer(read3).sequence == "CTATCGATCCA"

def test_TruSeq_trimmer():
    trimmer = TruSeqBisulfiteTrimmer()
    read1 = Sequence('read1', "CTATCGATCCACGAGACTAAC")
    assert trimmer(read1).sequence == "ATCCACGAGACTAAC"

def test_Swift_trimmer():
    trimmer = SwiftBisulfiteTrimmer()
    seq = "".join(["ACGT"] * 30)
    read1 = Sequence('read1', seq)
    read2 = Sequence('read2', seq)
    trimmed = trimmer(read1, read2)
    assert trimmed[0].sequence == seq[:-10]
    assert trimmed[1].sequence == seq[10:]
                            
def test_overlapping():
    trimmer = MergeOverlapping(min_overlap=10, error_rate = 0.1)
    a1 = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC'
    a2 = reverse_complement('AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA')
    frag = 'CCAAGCAGACATTCACTCAGATTGCA'
    r1 = (frag + a1)[0:40]
    q1 = '#' * 40
    r2 = reverse_complement(a2 + frag)[0:40]
    q2 = '!' * 40
    parser = AdapterParser()
    adapter1 = parser.parse(a1)
    adapter2 = parser.parse(a2)
    cutter = AdapterCutter([adapter1, adapter2])
    read1 = Sequence('foo', r1, q1)
    read1 = cutter(read1)
    assert len(read1) == 26
    read2 = Sequence('foo', r2, q2)
    read2 = cutter(read2)
    assert len(read2) == 26
    
    # complete overlap
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1 == read1_merged
    
    # partial overlap
    read1.merged = False
    read2 = read2.subseq(0, 24)[2]
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1 == read1_merged
    
    # partial overlap r1, r2
    read1.merged = False
    read1 = read1.subseq(0, 24)[2]
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert len(read1_merged) == 26
    assert read1_merged.sequence == 'CCAAGCAGACATTCACTCAGATTGCA'
    assert read1_merged.qualities == ('#' * 24) + ( '!' * 2)
    
    # errors
    # round(0.1 * 24) = 2, so 2 errors should pass but 3 should not
    read1.merged = False
    r1_seq = list(read1.sequence)
    r1_seq[10] = reverse_complement(r1_seq[10])
    r1_seq[20] = reverse_complement(r1_seq[20])
    read1.sequence = "".join(r1_seq)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert len(read1_merged) == 26
    assert read1_merged.sequence == 'CCAAGCAGACTTTCACTCAGTTTGCA'
    assert read1_merged.qualities == ('#' * 24) + ( '!' * 2)
    
    # too few overlapping bases
    read1.merged = False
    r1_seq[15] = reverse_complement(r1_seq[15])
    read1.sequence = "".join(r1_seq)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged is False
    assert read2 is not None

def test_overlapping_with_error_correction():
    trimmer = MergeOverlapping(min_overlap=10, error_rate = 0.1, mismatch_action='liberal')
    r1 = 'AGATCGGAAGACCGTCATGTAGGGAAAGAGTGTAGATCTC'
    q1 = 'FFFFFFFFFFF#FFFFFFFFFFFFFFFFFFFFF#######'
    r2 = reverse_complement('AGATCGGTAGAGCGTCGTGTAGGGAAATAGTGTAGATCTC')
    q2 = ''.join(reversed('FFFFFFFFFFFFFFFF#FFFFFFFFFF#FFFFFFFFFFFF'))
    read1 = Sequence('foo', r1, q1)
    read2 = Sequence('foo', r2, q2)
    read1_merged, read2_merged = trimmer(read1, read2)
    assert read1_merged.merged
    assert read2_merged is None
    assert read1_merged.sequence == 'AGATCGGTAGAGCGTCATGTAGGGAAAGAGTGTAGATCTC'
    assert read1_merged.qualities == 'FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF#######'

def test_mismatched_adapter_overlaps():
    """
    This is a test case from real data. The adapter overlaps 1 less bp
    on the fw read than on the reverse read. We want to make sure that
    the extra 'A' base gets trimmed.
    adapter                                                                                                                                                          GATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG
    actual                                                               TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGTAGATCGGAAGAGCACACGTCTGAACTTCAGTCAC
    actual rc                          TATGTTCTTTCCCTTCACGTCTCTCTTCGGATCTTTATTGTGATGAGTTGAAAATAAAGGTTAAGTATAGATAAAAAAGTTATTATAGTTTAGAGGGTAAGTGTATGATGGAGTAAAATATTGGT
    adapter rc AATGATACGGCGACCACCGAGATCTACACTCTTTCCCTACACGACGCTCTTCCGATCT
    """
    r1 = 'TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGTAGATCGGAAGAGCACACGTCTGAACTTCAGTCAC'
    r2 = 'ACCAATATTTTACTCCATCATACACTTACCCTCTAAACTATAATAACTTTTTTATCTATACTTAACCTTTATTTTCAACTCATCACAATAAAGATCCGAAGAGAGACGTGAAGGGAAAGAACATA'
    a1 = "GATCGGAAGAGCACACGTCTGAACTCCAGTCACCAGATCATCTCGTATGCCGTCTTCTGCTTG" # TruSeq index 7
    a2 = "AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT" # TruSeq universal

    parser = AdapterParser()
    adapter1 = parser.parse(a1)
    adapter2 = parser.parse(a2)
    # the data has a fairly high error rate
    cutter = InsertAdapterCutter(adapter1, adapter2,
        max_insert_mismatch_frac=0.3,
        max_adapter_mismatch_frac=0.3)
    read1 = Sequence('foo', r1, '#' * 125)
    read2 = Sequence('foo', r2, '#' * 125)
    new_read1, new_read2 = cutter(read1, read2)
    assert(len(new_read1)) == 91
    assert(len(new_read2)) == 91
    assert(new_read1.sequence == 'TTGTTTTTATGGAGAGAGTTTTAAGGTTTATTTTAGTTTTAAAGGATATTGTAGGTTAGAGGGAAAGTGTATGATGAAGGTATATATTGGT')
    
def test_error_correction():
    a1 = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC'
    a2 = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA'
    frag = 'CCAAGCAGACATTCACTCAGATTGCA'
    correct_frag = 'CCAAGTAGACATTCGCTCAGATTGCA'
    
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = 'T'
    q1 = ['#'] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = 'A'
    r1 = (''.join(r1) + a1)[0:40]
    q1 = ''.join(q1)
    
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = 'G'
    q2 = ['#'] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag)-15] = 'A'
    r2 = reverse_complement(reverse_complement(a2) + ''.join(r2))[0:40]
    q2 = ''.join(q2)
    
    read1 = Sequence('foo', r1, q1)
    read2 = Sequence('foo', r2, q2)
    
    parser = AdapterParser()
    adapter1 = parser.parse(a1)
    adapter2 = parser.parse(a2)
    
    cutter = InsertAdapterCutter(adapter1, adapter2, mismatch_action='liberal')
    
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)

def test_error_correction_no_insert_match_one_adapter_match():
    a1 = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC'
    a2 = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA'
    a2_mod = 'ACATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA'
    frag = 'CCAAGCAGACATTCACTCAGATTGCA'
    correct_frag = 'CCAAGTAGACATTCGCTCAGATTGCA'
    
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = 'T'
    q1 = ['#'] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = 'A'
    r1 = (''.join(r1) + a1)[0:40]
    q1 = ''.join(q1)
    
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = 'G'
    q2 = ['#'] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag)-15] = 'A'
    r2 = reverse_complement(reverse_complement(a2_mod) + ''.join(r2))[0:40]
    q2 = ''.join(q2)
    
    read1 = Sequence('foo', r1, q1)
    read2 = Sequence('foo', r2, q2)
    
    parser1 = AdapterParser()
    adapter1 = parser1.parse(a1)
    # Allow zero mismatches to prevent adapter alignment
    parser2 = AdapterParser(max_error_rate=0)
    adapter2 = parser2.parse(a2)
    
    # Allow zero mismatches to prevent insert alignment
    cutter = InsertAdapterCutter(
        adapter1, adapter2, mismatch_action='liberal',
        max_insert_mismatch_frac=0)
    
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert not new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert not new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)

def test_error_correction_no_insert_match_two_adapter_matches():
    a1 = 'AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTC'
    a2 = 'AGATCGGAAGAGCACACGTCTGAACTCCAGTCACGAGTTA'
    frag = 'CCAAGCAGACATTCACTCAGATTGCA'
    correct_frag = 'CCAAGTAGACATTCGCTCAGATTGCA'
    
    r1 = list(frag)
    # C>T at pos 6
    r1[5] = 'T'
    q1 = ['#'] * 40
    # quality of read1 > quality of read2 at pos 6
    q1[5] = 'A'
    r1 = (''.join(r1) + a1)[0:40]
    q1 = ''.join(q1)
    
    r2 = list(frag)
    # A>G at pos 15
    r2[14] = 'G'
    q2 = ['#'] * 40
    # quality of read2 > quality of read1 at pos 11
    q2[len(frag)-15] = 'A'
    r2 = reverse_complement(reverse_complement(a2) + ''.join(r2))[0:40]
    q2 = ''.join(q2)
    
    read1 = Sequence('foo', r1, q1)
    read2 = Sequence('foo', r2, q2)
    
    parser = AdapterParser()
    adapter1 = parser.parse(a1)
    adapter2 = parser.parse(a2)
    
    # Allow zero mismatches to prevent insert alignment
    cutter = InsertAdapterCutter(
        adapter1, adapter2, mismatch_action='liberal',
        max_insert_mismatch_frac=0)
    
    new_read1, new_read2 = cutter(read1, read2)
    assert len(new_read1) == 26
    assert not new_read1.insert_overlap
    assert new_read1.sequence == correct_frag
    assert len(new_read2) == 26
    assert not new_read2.insert_overlap
    assert new_read2.sequence == reverse_complement(correct_frag)

def ints2quals(ints):
    return ''.join(chr(i+33) for i in ints)

def test_overwrite_read():
    overwrite = OverwriteRead(20, 40, 10)
    lowseq = 'ACGT' * 5
    highseq = 'TCAG' * 5
    
    # mean lowq > 20, mean highq > 40
    lowq = (11, 31, 16, 24, 16, 20, 17, 19, 21, 28) * 2
    highq = (22, 62, 32, 48, 32, 40, 34, 38, 42, 56) * 2
    read1 = Sequence('foo', lowseq, ints2quals(lowq))
    read2 = Sequence('foo', highseq, ints2quals(highq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == lowseq
    assert new_read1.qualities == ints2quals(lowq)
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 0
    
    # mean lowq < 20, mean highq > 40
    lowq = tuple(i-1 for i in lowq)
    read1 = Sequence('foo', lowseq, ints2quals(lowq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == rc(highseq)
    assert new_read1.qualities == ints2quals(reversed(highq))
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 1
    
    # mean lowq < 20, mean highq < 40
    highq = tuple(i-1 for i in highq)
    read2 = Sequence('foo', highseq, ints2quals(highq))
    new_read1, new_read2 = overwrite(read1, read2)
    assert new_read1.sequence == lowseq
    assert new_read1.qualities == ints2quals(lowq)
    assert new_read2.sequence == highseq
    assert new_read2.qualities == ints2quals(highq)
    assert new_read1.corrected == new_read2.corrected == 0



# coding: utf-8
"""Sequence I/O classes: Reading and writing of FASTA and FASTQ files.

TODO
- Sequence.name should be Sequence.description or so (reserve .name for the part
  before the first space)
"""
import sys
from atropos import AtroposError
from atropos.io import STDOUT, xopen
from atropos.io.compression import splitext_compressed
from atropos.util import Summarizable, truncate_string

READ1 = 1
READ2 = 2
SINGLE = READ1
PAIRED = 1|2

class FormatError(AtroposError):
    """Raised when an input file (FASTA or FASTQ) is malformatted."""
    pass

class UnknownFileType(AtroposError):
    """Raised when open could not autodetect the file type."""
    pass

## Reading sequences from files ##

class SequenceReaderBase(Summarizable): # pylint: disable=no-member
    """Sequence readers must provide the following properties:
    - input_names: (read1 file, read2 file)
    - input_read: 1, 2, or 3
    - file_format: string, e.g. FASTA, FASTQ, etc.
    - delivers_qualities: bool
    - has_qualfile: bool
    - quality_base: int
    - colorspace: bool
    - interleaved: bool
    """
    def summarize(self):
        return dict(
            input_names=self.input_names,
            input_read=self.input_read,
            file_format=self.file_format,
            delivers_qualities=self.delivers_qualities,
            quality_base=self.quality_base,
            has_qualfile=self.has_qualfile,
            colorspace=self.colorspace,
            interleaved=self.interleaved)

class SequenceReader(SequenceReaderBase):
    """Read possibly compressed files containing sequences.
    
    Args:
        file is a path or a file-like object. In both cases, the file may
            be compressed (.gz, .bz2, .xz).
        mode: The file open mode.
    """
    delivers_qualities = False
    has_qualfile = False
    colorspace = False
    interleaved = False
    input_read = SINGLE
    _close_on_exit = False
    
    def __init__(self, path, mode='r', quality_base=None):
        if isinstance(path, str):
            self.name = path
            self.quality_base = quality_base
            self._file = xopen(path, mode)
            self._close_on_exit = True
        else:
            if hasattr(path, 'name'):
                self.name = path.name
            else:
                # TODO: generate random unique name?
                self.name = path.__class__
            self._file = path
    
    @property
    def input_names(self):
        return (self.name, None)
    
    def close(self):
        """Close the underlying file.
        """
        if self._close_on_exit and self._file is not None:
            self._file.close()
            self._file = None
    
    def __enter__(self):
        if self._file is None:
            raise ValueError("I/O operation on closed SequenceReader")
        return self
    
    def __exit__(self, *args):
        self.close()

try:
    from ._seqio import Sequence, FastqReader
except ImportError:
    pass

class ColorspaceSequence(Sequence):
    """Sequence object for colorspace reads.
    """
    def __init__(
            self, name, sequence, qualities, primer=None, name2='',
            original_length=None, match=None, match_info=None, clipped=None,
            insert_overlap=False, merged=False, corrected=0):
        # In colorspace, the first character is the last nucleotide of the
        # primer base and the second character encodes the transition from the
        # primer base to the first real base of the read.
        if primer is None:
            self.primer = sequence[0:1]
            sequence = sequence[1:]
        else:
            self.primer = primer
        if qualities is not None and len(sequence) != len(qualities):
            rname = truncate_string(name)
            raise FormatError(
                "In read named {0!r}: length of colorspace quality "
                "sequence ({1}) and length of read ({2}) do not match (primer "
                "is: {3!r})".format(
                    rname, len(qualities), len(sequence), self.primer))
        super().__init__(
            name, sequence, qualities, name2, original_length, match,
            match_info, clipped, insert_overlap, merged, corrected)
        if not self.primer in ('A', 'C', 'G', 'T'):
            raise FormatError(
                "Primer base is {0!r} in read {1!r}, but it should be one of "
                "A, C, G, T.".format(self.primer, truncate_string(name)))
    
    def __repr__(self):
        fmt_str = \
            '<ColorspaceSequence(name={0!r}, primer={1!r}, sequence={2!r}{3})>'
        qstr = ''
        if self.qualities is not None:
            qstr = ', qualities={0!r}'.format(truncate_string(self.qualities))
        return fmt_str.format(
            truncate_string(self.name), self.primer,
            truncate_string(self.sequence), qstr)
    
    def __getitem__(self, key):
        return self.__class__(
            self.name,
            self.sequence[key],
            self.qualities[key] if self.qualities is not None else None,
            self.primer,
            self.name2,
            self.original_length,
            self.match,
            self.match_info,
            self.clipped,
            self.insert_overlap,
            self.merged,
            self.corrected)

class FileWithPrependedLine(object):
    """A file-like object that allows to "prepend" a single line to an already
    opened file. That is, further reads on the file will return the provided
    line and only then the actual content. This is needed to solve the problem
    of autodetecting input from a stream: As soon as the first line has been
    read, we know the file type, but also that line is "gone" and unavailable
    for further processing.
    
    Args:
        file: An already opened file-like object.
        line: A single string (newline will be appended if not included).
    """
    def __init__(self, file, line):
        if not line.endswith('\n'):
            line += '\n'
        self.first_line = line
        self._file = file
    
    @property
    def name(self):
        return self._file.name
    
    def __iter__(self):
        yield self.first_line
        for line in self._file:
            yield line
    
    def close(self):
        """Close the underlying file.
        """
        self._file.close()

class FastaReader(SequenceReader):
    """Reader for FASTA files.
    
    Args:
        path: A path or a file-like object. In both cases, the file may
            be compressed (.gz, .bz2, .xz).
        keep_linebreaks: Whether to keep newline characters in the sequence.
        sequence_class: The class to use when creating new sequence objects.
    """
    file_format = "FASTA"
    
    def __init__(self, path, keep_linebreaks=False, sequence_class=Sequence):
        super().__init__(path)
        self.sequence_class = sequence_class
        self._delimiter = '\n' if keep_linebreaks else ''
    
    def __iter__(self):
        """Read next entry from the file (single entry at a time).
        """
        name = None
        seq = []
        for i, line in enumerate(self._file):
            # strip() also removes DOS line breaks
            line = line.strip()
            if not line:
                continue
            if line and line[0] == '>':
                if name is not None:
                    yield self.sequence_class(
                        name, self._delimiter.join(seq), None)
                name = line[1:]
                seq = []
            elif line and line[0] == '#':
                continue
            elif name is not None:
                seq.append(line)
            else:
                raise FormatError(
                    "At line {0}: Expected '>' at beginning of FASTA record, "
                    "but got {1!r}.".format(i+1, truncate_string(line)))
        
        if name is not None:
            yield self.sequence_class(name, self._delimiter.join(seq), None)

class ColorspaceFastaReader(FastaReader):
    """Reads colorspace sequences from a FASTA.
    
    Args:
        path: The file to read.
        keep_linebreaks: Whether to keep linebreaks in wrapped sequences.
    """
    colorspace = True
    
    def __init__(self, path, keep_linebreaks=False):
        super().__init__(
            path, keep_linebreaks, sequence_class=ColorspaceSequence)

class ColorspaceFastqReader(FastqReader):
    """Reads colorspace sequences from a FASTQ.
    """
    colorspace = True
    
    def __init__(self, path, quality_base=33):
        super().__init__(
            path, quality_base=quality_base, sequence_class=ColorspaceSequence)

class SRAColorspaceFastqReader(FastqReader):
    """Reads SRA-formatted colorspace sequences from a FASTQ.
    """
    colorspace = True
    
    def __init__(self, path, quality_base=33):
        super().__init__(
            path, quality_base=quality_base,
            sequence_class=sra_colorspace_sequence)

class FastaQualReader(SequenceReaderBase):
    """Reader for reads that are stored in .(CS)FASTA and .QUAL files.
    
    Args:
        fastafile and qualfile are filenames or file-like objects.
            If a filename is used, then .gz files are recognized.
        sequence_class: The class to use when creating new sequence objects.
    """
    file_format = "FastaQual"
    delivers_qualities = True
    has_qualfile = True
    colorspace = False
    interleaved = False
    input_read = SINGLE
    
    def __init__(
            self, fastafile, qualfile, quality_base=33,
            sequence_class=Sequence):
        self.fastareader = FastaReader(fastafile)
        self.qualreader = FastaReader(qualfile, keep_linebreaks=True)
        self.quality_base = quality_base
        self.sequence_class = sequence_class
    
    @property
    def input_names(self):
        return ((self.fastareader.name, self.qualreader.name), None)
    
    def __iter__(self):
        """Yield Sequence objects.
        """
        # conversion dictionary: maps strings to the appropriate ASCII-encoded
        # character
        conv = dict()
        for i in range(-5, 256 - 33):
            conv[str(i)] = chr(i + 33)
        for fastaread, qualread in zip(self.fastareader, self.qualreader):
            if fastaread.name != qualread.name:
                raise FormatError(
                    "The read names in the FASTA and QUAL file do not match "
                    "({0!r} != {1!r})".format(fastaread.name, qualread.name))
            try:
                qualities = ''.join(
                    [conv[value] for value in qualread.sequence.split()])
            except KeyError as err:
                raise FormatError(
                    "Within read named {0!r}: Found invalid quality "
                    "value {1}".format(fastaread.name, err))
            assert fastaread.name == qualread.name
            yield self.sequence_class(
                fastaread.name, fastaread.sequence, qualities)
    
    def close(self):
        """Close the underlying files.
        """
        self.fastareader.close()
        self.qualreader.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()

class ColorspaceFastaQualReader(FastaQualReader):
    """Reads sequences and qualities from separate files and returns
    :class:`ColorspaceSequence`s.
    
    Args:
        fastafile, qualfile: FASTA files that contain the sequences and
            qualities, respectively.
    """
    colorspace = True
    
    def __init__(self, fastafile, qualfile, quality_base=33):
        super().__init__(
            fastafile, qualfile, quality_base=quality_base,
            sequence_class=ColorspaceSequence)

class PairedSequenceReader(SequenceReaderBase):
    """Read paired-end reads from two files. Wraps two SequenceReader instances,
    making sure that reads are properly paired.
    
    Args:
        file1, file2: The pair of files.
        colorspace: Whether the sequences are in colorspace.
        file_format: A file_format instance.
    """
    input_read = PAIRED
    interleaved = False
    
    def __init__(
            self, file1, file2, quality_base=33, colorspace=False,
            file_format=None):
        self.reader1 = open_reader(
            file1, colorspace=colorspace, quality_base=quality_base,
            file_format=file_format)
        self.reader2 = open_reader(
            file2, colorspace=colorspace, quality_base=quality_base,
            file_format=file_format)
    
    @property
    def input_names(self):
        return (
            self.reader1.input_names[0], 
            self.reader2.input_names[0])
    
    def __getattr__(self, name):
        return getattr(self.reader1, name)
    
    def __iter__(self):
        """Iterate over the paired reads. Each item is a pair of Sequence
        objects.
        """
        # Avoid usage of zip() below since it will consume one item too many.
        it1, it2 = iter(self.reader1), iter(self.reader2)
        while True:
            try:
                read1 = next(it1)
            except StopIteration:
                # End of file 1. Make sure that file 2 is also at end.
                try:
                    next(it2)
                    raise FormatError(
                        "Reads are improperly paired. There are more reads in "
                        "file 2 than in file 1.")
                except StopIteration:
                    pass
                break
            try:
                read2 = next(it2)
            except StopIteration:
                raise FormatError(
                    "Reads are improperly paired. There are more reads in "
                    "file 1 than in file 2.")
            if not sequence_names_match(read1, read2):
                raise FormatError(
                    "Reads are improperly paired. Read name '{0}' in file 1 "
                    "does not match '{1}' in file 2.".format(
                        read1.name, read2.name))
            yield (read1, read2)
    
    def close(self):
        """Close the underlying files.
        """
        self.reader1.close()
        self.reader2.close()
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()

class InterleavedSequenceReader(SequenceReaderBase):
    """Read paired-end reads from an interleaved FASTQ file.
    
    Args:
        path: The interleaved FASTQ file.
        colorspace: Whether the sequences are in colorspace.
        file_format: A file_format instance.
    """
    input_read = PAIRED
    interleaved = True
    
    def __init__(
            self, path, quality_base=33, colorspace=False, file_format=None):
        self.reader = open_reader(
            path, quality_base=quality_base, colorspace=colorspace,
            file_format=file_format)
    
    def __getattr__(self, name):
        return getattr(self.reader, name)
    
    def __iter__(self):
        # Avoid usage of zip() below since it will consume one item too many.
        itr = iter(self.reader)
        for read1 in itr:
            try:
                read2 = next(itr)
            except StopIteration:
                raise FormatError(
                    "Interleaved input file incomplete: Last record has no "
                    "partner.")
            if not sequence_names_match(read1, read2):
                raise FormatError(
                    "Reads are improperly paired. Name {0!r} (first) does not "
                    "match {1!r} (second).".format(read1.name, read2.name))
            yield (read1, read2)
    
    def close(self):
        """Close the underlying reader.
        """
        self.reader.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

# TODO: SAM/BAM classes need unit tests

class SAMReader(SequenceReaderBase):
    """Reader for SAM/BAM files. Paired-end files must be name-sorted. Does
    not support secondary/supplementary reads. This is an abstract class.
    
    Args:
        path: A filename or a file-like object. If a filename, then .gz files
            are supported.
        sequence_class: The class to use when creating new sequence objects.
    """
    file_format = "SAM"
    delivers_qualities = True
    interleaved = False
    has_qualfile = False
    colorspace = False
    
    def __init__(self, path, quality_base=33, sequence_class=Sequence):
        self._close_on_exit = False
        if isinstance(path, str):
            path = xopen(path, 'rb')
            self._close_on_exit = True
        if isinstance(path, str):
            self.name = path
        else:
            self.name = path.name
        self._file = path
        self.quality_base = quality_base
        self.sequence_class = sequence_class
    
    @property
    def input_names(self):
        return (self.name, None)
    
    def __iter__(self):
        import pysam
        return self._iter(pysam.AlignmentFile(self._file))
    
    def _iter(self, sam):
        """Create an iterator over records in the SAM/BAM file.
        """
        raise NotImplementedError()
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self.close()
    
    def close(self):
        """Close the underling AlignmentFile.
        """
        if self._close_on_exit and self._file is not None:
            self._file.close()
            self._file = None
        
    def _as_sequence(self, read):
        return self.sequence_class(
            read.query_name,
            read.query_sequence,
            ''.join(chr(33 + q) for q in read.query_qualities))

class SingleEndSAMReader(SAMReader):
    """Reader for single-end SAM/BAM files.
    """
    input_read = SINGLE
    
    def _iter(self, sam):
        for read in sam:
            yield self._as_sequence(read)

class Read1SingleEndSAMReader(SAMReader):
    """Reads a paired-end SAM/BAM file as if it were single-end, yielding
    only the first read from each pair.
    """
    input_read = READ1
    
    def _iter(self, sam):
        for read in sam:
            if read.is_read1:
                yield self._as_sequence(read)

class Read2SingleEndSAMReader(SAMReader):
    """Reads a paired-end SAM/BAM file as if it were single-end, yielding
    only the second read from each pair.
    """
    input_read = READ2
    
    def _iter(self, sam):
        for read in sam:
            if read.is_read2:
                yield self._as_sequence(read)

class PairedEndSAMReader(SAMReader):
    """Reads pairs of reads from a SAM/BAM file. The file must be name-sorted.
    """
    input_read = PAIRED
    interleaved = True
    
    def _iter(self, sam):
        for reads in zip(sam, sam):
            if reads[0].query_name != reads[1].query_name:
                raise AtroposError(
                    "Consecutive reads {}, {} in paired-end SAM/BAM file do "
                    "not have the same name; make sure your file is "
                    "name-sorted and does not contain any "
                    "secondary/supplementary alignments.",
                    reads[0].query_name, reads[1].query_name)
            
            if reads[0].is_read1:
                assert reads[1].is_read2
            else:
                assert reads[1].is_read1
                reads = (reads[1], reads[0])
            
            yield tuple(self._as_sequence(r) for r in reads)

class SequenceFileFormat(object):
    """Base class for sequence formatters.
    """
    def format(self, read):
        """Format a Sequence as a string.
        
        Args:
            read: The Sequence object.
        
        Returns:
            A string representation of the sequence object in the sequence
            file format.
        """
        raise NotImplementedError()

class FastaFormat(SequenceFileFormat):
    """FASTA SequenceFileFormat.
    
    Args:
        line_length: Max line length (in characters), or None. Determines
            whether and how lines are wrapped.
    """
    def __init__(self, line_length=None):
        self.text_wrapper = None
        if line_length:
            from textwrap import TextWrapper
            self.text_wrapper = TextWrapper(width=line_length)
    
    def format(self, read):
        return self.format_entry(read.name, read.sequence)
    
    def format_entry(self, name, sequence):
        """Convert a sequence record to a string.
        """
        if self.text_wrapper:
            sequence = self.text_wrapper.fill(sequence)
        return "".join((">", name, "\n", sequence, "\n"))

class ColorspaceFastaFormat(FastaFormat):
    """FastaFormat in which sequences are in colorspace.
    """
    def format(self, read):
        return self.format_entry(read.name, read.primer + read.sequence)

class FastqFormat(SequenceFileFormat):
    """FASTQ SequenceFileFormat.
    """
    def format(self, read):
        return self.format_entry(
            read.name, read.sequence, read.qualities, read.name2)
    
    def format_entry(self, name, sequence, qualities, name2=""):
        """Convert a sequence record to a string.
        """
        return "".join((
            '@', name, '\n',
            sequence, '\n+',
            name2, '\n',
            qualities, '\n'))

class ColorspaceFastqFormat(FastqFormat):
    """FastqFormat in which sequences are in colorspace.
    """
    def format(self, read):
        return self.format_entry(
            read.name, read.primer + read.sequence, read.qualities)

class SingleEndFormatter(object):
    """Wrapper for a SequenceFileFormat for single-end data.
    
    Args:
        seq_format: The SequenceFileFormat object.
        file1: The single-end file.
    """
    def __init__(self, seq_format, file1):
        self.seq_format = seq_format
        self.file1 = file1
        self.written = 0
        self.read1_bp = 0
        self.read2_bp = 0
    
    def format(self, result, read1, read2=None):
        """Format read(s) and add them to `result`.
        
        Args:
            result: A dict mapping file names to lists of formatted reads.
            read1, read2: The reads to format.
        """
        result[self.file1].append(self.seq_format.format(read1))
        self.written += 1
        self.read1_bp += len(read1)
    
    @property
    def written_bp(self):
        """Tuple of base-pairs written (read1_bp, read2_bp).
        """
        return (self.read1_bp, self.read2_bp)

class InterleavedFormatter(SingleEndFormatter):
    """Format read pairs as successive reads in an interleaved file.
    """
    def format(self, result, read1, read2=None):
        result[self.file1].extend((
            self.seq_format.format(read1),
            self.seq_format.format(read2)))
        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)

class PairedEndFormatter(SingleEndFormatter):
    """Wrapper for a SequenceFileFormat. Both reads in a pair are formatted
    using the specified format.
    """
    def __init__(self, seq_format, file1, file2):
        super(PairedEndFormatter, self).__init__(seq_format, file1)
        self.file2 = file2
    
    def format(self, result, read1, read2):
        result[self.file1].append(self.seq_format.format(read1))
        result[self.file2].append(self.seq_format.format(read2))
        self.written += 1
        self.read1_bp += len(read1)
        self.read2_bp += len(read2)

def sra_colorspace_sequence(name, sequence, qualities, name2):
    """Factory for an SRA colorspace sequence (which has one quality value
    too many).
    """
    return ColorspaceSequence(name, sequence, qualities[1:], name2=name2)

def sequence_names_match(read1, read2):
    """Check whether the sequences read1 and read2 have identical names,
    ignoring a suffix of '1' or '2'. Some old paired-end reads have names that
    end in '/1' and '/2'. Also, the fastq-dump tool (used for converting SRA
    files to FASTQ) appends a .1 and .2 to paired-end reads if option -I is
    used.
    
    Args:
        read1, read2: The sequences to compare.
    
    Returns:
        Whether the sequences are equal.
    """
    name1 = read1.name.split(None, 1)[0]
    name2 = read2.name.split(None, 1)[0]
    if name1[-1:] in '12' and name2[-1:] in '12':
        name1 = name1[:-1]
        name2 = name2[:-1]
    return name1 == name2

def paired_to_read1(reader):
    """Generator that yields the first read from an iterator over read pairs.
    """
    for read1, _ in reader:
        yield read1

def paired_to_read2(reader):
    """Generator that yields the second read from an iterator over read pairs.
    """
    for _, read2 in reader:
        yield read2

def open_reader(
        file1, file2=None, qualfile=None, quality_base=None, colorspace=False,
        file_format=None, interleaved=False, input_read=None):
    """Open sequence files in FASTA or FASTQ format for reading. This is
    a factory that returns an instance of one of the ...Reader
    classes also defined in this module.
    
    Args:
        file1, file2, qualfile: Paths to regular or compressed files or
            file-like objects. Use file1 if data is single-end. If file2 is also
            provided, sequences are paired. If qualfile is given, then file1
            must be a FASTA file and sequences are single-end. One of file2 and
            qualfile must always be None (no paired-end data is supported when
            reading qualfiles).
        quality_base: Base for quality values.
        interleaved:If True, then file1 contains interleaved paired-end data.
            file2 and qualfile must be None in this case.
        colorspace: If True, instances of the Colorspace... classes
            are returned.
        file_format: If set to None, file format is autodetected from the file
            name extension. Set to 'fasta', 'fastq', 'sra-fastq', 'sam', or
            'bam' to not auto-detect. Colorspace is not auto-detected and must
            always be requested explicitly.
        input_read: When file1 is a paired-end interleaved or SAM/BAM
            file, this specifies whether to only use the first or second read
            (1 or 2) or to use both reads (None).
    """
    if interleaved and (file2 is not None or qualfile is not None):
        raise ValueError(
            "When interleaved is set, file2 and qualfile must be None")
    if file2 is not None and qualfile is not None:
        raise ValueError("Setting both file2 and qualfile is not supported")
    
    if file2 is not None:
        return PairedSequenceReader(
            file1, file2, quality_base=quality_base, colorspace=colorspace,
            file_format=file_format)
    
    if qualfile is not None:
        if colorspace:
            # read from .(CS)FASTA/.QUAL
            return ColorspaceFastaQualReader(
                file1, qualfile, quality_base=quality_base)
        else:
            return FastaQualReader(
                file1, qualfile, quality_base=quality_base)
    
    if file_format is None and file1 != STDOUT:
        file_format = guess_format_from_name(file1)
    
    if file_format is None:
        if file1 == STDOUT:
            file1 = sys.stdin
        for line in file1:
            if line.startswith('#'):
                # Skip comment lines (needed for csfasta)
                continue
            if line.startswith('>'):
                file_format = 'fasta'
            elif line.startswith('@'):
                file_format = 'fastq'
            # TODO: guess SAM/BAM from data
            file1 = FileWithPrependedLine(file1, line)
            break
    
    if file_format is not None:
        file_format = file_format.lower()
        if file_format in ("sam", "bam"):
            if colorspace:
                raise ValueError(
                    "SAM/BAM format is not currently supported for colorspace "
                    "reads")
            if interleaved:
                return PairedEndSAMReader(file1, quality_base=quality_base)
            elif input_read == READ1:
                return Read1SingleEndSAMReader(file1, quality_base=quality_base)
            elif input_read == READ2:
                return Read2SingleEndSAMReader(file1, quality_base=quality_base)
            else:
                return SingleEndSAMReader(file1, quality_base=quality_base)
        elif interleaved:
            reader = InterleavedSequenceReader(
                file1, quality_base=quality_base, colorspace=colorspace,
                file_format=file_format)
            if input_read == READ1:
                return paired_to_read1(reader)
            elif input_read == READ2:
                return paired_to_read2(reader)
            else:
                return reader
        elif file_format == 'fasta':
            fasta_handler = ColorspaceFastaReader if colorspace else FastaReader
            return fasta_handler(file1)
        elif file_format == 'fastq':
            fastq_handler = ColorspaceFastqReader if colorspace else FastqReader
            return fastq_handler(file1, quality_base=quality_base)
        elif file_format == 'sra-fastq' and colorspace:
            return SRAColorspaceFastqReader(file1, quality_base=quality_base)
    
    raise UnknownFileType(
        "File format {0!r} is unknown (expected 'sra-fastq' (only for "
        "colorspace), 'fasta', 'fastq', 'sam', or 'bam').".format(file_format))

def guess_format_from_name(path, raise_on_failure=False):
    """Detect file format based on the file name.
    
    Args:
        path: The filename to guess.
        raise_on_failure: Whether to raise an exception if the filename cannot
            be detected.
    
    Returns:
        The format name.
    """
    name = None
    if isinstance(path, str):
        name = path
    elif hasattr(path, "name"):	 # seems to be an open file-like object
        name = path.name
    
    if name:
        name, ext1, _ = splitext_compressed(name)
        ext = ext1.lower()
        if ext in ['.fasta', '.fa', '.fna', '.csfasta', '.csfa']:
            return 'fasta'
        elif ext in ['.fastq', '.fq'] or (
                ext == '.txt' and name.endswith('_sequence')):
            return 'fastq'
        elif ext in ('.sam', '.bam'):
            return ext[1:]
    
    if raise_on_failure:
        raise UnknownFileType(
            "Could not determine whether file {0!r} is FASTA or FASTQ: file "
            "name extension {1!r} not recognized".format(path, ext))

def create_seq_formatter(file1, file2=None, interleaved=False, **kwargs):
    """Create a formatter, deriving the format name from the file extension.
    
    Args:
        file1, file2: Output files.
        interleaved: Whether the output should be interleaved (file2 must be
            None).
        kwargs: Additional arguments to pass to :method:`get_format`.
    """
    seq_format = get_format(file1, **kwargs)
    if file2 is not None:
        return PairedEndFormatter(seq_format, file1, file2)
    elif interleaved:
        return InterleavedFormatter(seq_format, file1)
    else:
        return SingleEndFormatter(seq_format, file1)

def get_format(
        path, file_format=None, colorspace=False, qualities=None,
        line_length=None):
    """Create a SequenceFileFormat instance.
    
    Args:
        path: The filename.
        
        file_format: If set to None, file format is autodetected from the file
        name extension. Set to 'fasta', 'fastq', or 'sra-fastq' to not
        auto-detect. Colorspace is not auto-detected and must always be
        requested explicitly.
        
        colorspace: If True, instances of the Colorspace... formats are
            returned.
        
        qualities: When file_format is None, this can be set to True or False to
            specify whether the written sequences will have quality values.
            This is is used in two ways:
            * If the output format cannot be determined (unrecognized extension
              etc), no exception is raised, but fasta or fastq format is chosen
              appropriately.
            * When False (no qualities available), an exception is raised when
              the auto-detected output format is FASTQ.
    
    Returns:
        A SequenceFileFormat object.
    """
    if file_format is None:
        file_format = guess_format_from_name(
            path, raise_on_failure=qualities is None)
    
    if file_format is None:
        if qualities is True:
            # Format not recognized, but know we want to write reads with
            # qualities.
            file_format = 'fastq'
        elif qualities is False:
            # Same, but we know that we want to write reads without qualities.
            file_format = 'fasta'
    
    if file_format is None:
        raise UnknownFileType("Could not determine file type.")
    
    if file_format == 'fastq' and qualities is False:
        raise ValueError(
            "Output format cannot be FASTQ since no quality values are "
            "available.")
    
    file_format = file_format.lower()
    if file_format == 'fasta':
        if colorspace:
            return ColorspaceFastaFormat(line_length)
        else:
            return FastaFormat(line_length)
    elif file_format == 'fastq':
        if colorspace:
            return ColorspaceFastqFormat()
        else:
            return FastqFormat()
    else:
        raise UnknownFileType(
            "File format {0!r} is unknown (expected 'fasta' or "
            "'fastq').".format(file_format))



"""Command-line interface for the detect command.
"""
from atropos.io import STDOUT
from atropos.commands.cli import (
    BaseCommandParser, positive, readable_url, writeable_file,
    readwriteable_file)

class CommandParser(BaseCommandParser):
    name = 'detect'
    usage = """
atropos detect -se input.fastq
atropos detect -pe1 in1.fq -pe2 in2.fq
"""
    description = """
Detect adapter sequences directly from read sequences. Use this command if you
are unsure if your data has adapters, or if you know that it has adapters but
you don't know what are the adapter sequences.
"""
    
    def add_command_options(self):
        parser = self.parser
        parser.set_defaults(
            max_reads=10000,
            counter_magnitude="K",
            report_formats=['txt'])
        
        group = self.add_group("Adapter Detection")
        group.add_argument(
            "-d",
            "--detector",
            choices=('known', 'heuristic', 'khmer'), default=None,
            help="Which detector to use. (automatically choose based on other "
                 "options)")
        group.add_argument(
            "-k",
            "--kmer-size",
            type=positive(), default=12,
            help="Size of k-mer used to scan reads for adapter sequences. (12)")
        group.add_argument(
            "-i",
            "--include-contaminants",
            choices=('all','known','unknown'), default='all',
            help="What conaminants to search for: all, only known "
                 "adapters/contaminants ('known'), or only unknown "
                 "contaminants ('unknown'). (all)")
        group.add_argument(
            "-x",
            "--known-contaminant",
            action="append", dest='known_adapter', default=None,
            help="Pass known contaminants in on the commandline as "
                 "'name=sequence'. Can be specified multiple times.")
        group.add_argument(
            "-F",
            "--known-contaminants-file",
            type=readable_url, action="append", dest='known_adapters_file',
            default=None,
            help="Points to FASTA File or URL with known contaminants.")
        group.add_argument(
            "--no-default-contaminants",
            action="store_false", dest="default_adapters", default=True,
            help="Don't fetch the default contaminant list (which is currently "
                 "stored as a GitHub gist).")
        group.add_argument(
            "--contaminant-cache-file",
            type=readwriteable_file, dest='adapter_cache_file',
            default='.adapters',
            help="File where known contaminant sequences will be cached, "
                 "unless --no-cache-contaminants is set.")
        group.add_argument(
            "--no-cache-contaminants",
            action="store_false", dest="cache_adapters", default=True,
            help="Don't cache contaminant list as '.contaminants' in the "
                 "working directory.")
        
        group = self.add_group("Output")
        group.add_argument(
            "-o",
            "--output",
            type=writeable_file, default=STDOUT, metavar="FILE",
            help="File in which to write the summary of detected adapters. "
                 "(stdout)")
        group.add_argument(
            "-m",
            "--max-adapters",
            type=positive(), default=None,
            help="The maximum number of candidate adapters to report. "
                 "(report all)")
    
    def validate_command_options(self, options):
        options.report_file = options.output



# kate: syntax Python;
# cython: profile=False, emit_code_comments=False
import copy
from atropos.io import xopen
from atropos.io.seqio import FormatError, SequenceReader
from atropos.util import reverse_complement, truncate_string

cdef class Sequence(object):
    """
    A record in a FASTQ file. Also used for FASTA (then the qualities attribute
    is None). qualities is a string and it contains the qualities encoded as
    ascii(qual+33).

    If an adapter has been matched to the sequence, the 'match' attribute is
    set to the corresponding Match instance.
    """
    cdef:
        public str name
        public str sequence
        public str qualities
        public str name2
        public int original_length
        public object match
        public object match_info
        public object clipped
        public bint insert_overlap
        public bint merged
        public int corrected
    
    def __init__(self, str name, str sequence, str qualities=None, str name2='',
                 original_length=None, match=None, match_info=None, clipped=None,
                 insert_overlap=False, merged=False, corrected=0):
        """Set qualities to None if there are no quality values"""
        self.name = name
        self.sequence = sequence
        self.qualities = qualities
        self.name2 = name2
        self.original_length = original_length or len(sequence)
        self.match = match
        self.match_info = match_info
        self.clipped = clipped or [0,0,0,0]
        self.insert_overlap = insert_overlap
        self.merged = merged
        self.corrected = corrected
        if qualities is not None and len(qualities) != len(sequence):
            rname = truncate_string(name)
            raise FormatError(
                "In read named {0!r}: length of quality sequence ({1}) and "
                "length  of read ({2}) do not match".format(
                    rname, len(qualities), len(sequence)))
    
    def subseq(self, begin=0, end=None):
        if end is None:
            new_read = self[begin:]
        else:
            new_read = self[begin:end]
        end_bases = len(self) - end
        offset = 2 if self.match else 0
        if begin:
            new_read.clipped[offset] += begin
        if end_bases:
            new_read.clipped[offset+1] += end_bases
        return (begin, end_bases, new_read)
    
    def clip(self, front=0, back=0):
        if back < 0:
            new_read = self[front:back]
            back *= -1
        else:
            new_read = self[front:]
        offset = 2 if self.match else 0
        if front:
            new_read.clipped[offset] += front
        if back:
            new_read.clipped[offset+1] += back
        return (front, back, new_read)
    
    def reverse_complement(self):
        """Returns a copy of this read with the sequence reverse-complemented
        and the qualities reversed.
        
        TODO: add option to also reverse the alignment positions
        """
        sequence = reverse_complement(self.sequence)
        qualities = clipped = match_info = None
        if self.qualities:
            qualities = ''.join(reversed(self.qualities))
        if self.match_info:
            match_info = [copy.copy(m) for m in self.match_info]
        
        new_read = self.__class__(
            self.name,
            sequence,
            qualities,
            self.name2,
            self.original_length,
            None,
            match_info,
            list(self.clipped),
            self.insert_overlap,
            self.merged,
            self.corrected
        )
        
        if self.match:
            match = self.match.copy()
            match.read = new_read
            new_read.match = match
        
        return new_read
    
    def __getitem__(self, key):
        """slicing"""
        return self.__class__(
            self.name,
            self.sequence[key],
            self.qualities[key] if self.qualities is not None else None,
            self.name2,
            self.original_length,
            self.match,
            self.match_info,
            list(self.clipped),
            self.insert_overlap,
            self.merged,
            self.corrected
        )

    def __repr__(self):
        qstr = ''
        if self.qualities is not None:
            qstr = ', qualities={0!r}'.format(truncate_string(self.qualities))
        return '<Sequence(name={0!r}, sequence={1!r}{2})>'.format(
            truncate_string(self.name), truncate_string(self.sequence), qstr)

    def __len__(self):
        return len(self.sequence)

    def __richcmp__(self, other, int op):
        if 2 <= op <= 3:
            eq = self.name == other.name and \
                self.sequence == other.sequence and \
                self.qualities == other.qualities
            if op == 2:
                return eq
            else:
                return not eq
        else:
            raise NotImplementedError()

    def __reduce__(self):
        return (Sequence, (self.name, self.sequence, self.qualities, self.name2))

class FastqReader(SequenceReader):
    """Reader for FASTQ files. Does not support multi-line FASTQ files.
    """
    file_format = "FASTQ"
    delivers_qualities = True
    
    def __init__(self, filename, quality_base=33, sequence_class=Sequence):
        """
        file is a filename or a file-like object.
        If file is a filename, then .gz files are supported.
        """
        super().__init__(filename, quality_base=quality_base)
        self.sequence_class = sequence_class
    
    def __iter__(self):
        """
        Yield Sequence objects
        """
        cdef int i = 0
        cdef int strip
        cdef str line, name, qualities, sequence, name2
        sequence_class = self.sequence_class

        it = iter(self._file)
        line = next(it)
        if not (line and line[0] == '@'):
            raise FormatError("Line {0} in FASTQ file is expected to start "
                              "with '@', but found {1!r}".format(i+1, line[:10]))
        strip = -2 if line.endswith('\r\n') else -1
        name = line[1:strip]

        i = 1
        for line in it:
            if i == 0:
                if not (line and line[0] == '@'):
                    raise FormatError("Line {0} in FASTQ file is expected to "
                                      "start with '@', but found {1!r}".format(
                                      i+1, line[:10]))
                name = line[1:strip]
            elif i == 1:
                sequence = line[:strip]
            elif i == 2:
                if line == '+\n':  # check most common case first
                    name2 = ''
                else:
                    line = line[:strip]
                    if not (line and line[0] == '+'):
                        raise FormatError("Line {0} in FASTQ file is expected "
                                          "to start with '+', but found {1!r}".format(
                                          i+1, line[:10]))
                    if len(line) > 1:
                        if not line[1:] == name:
                            raise FormatError(
                                "At line {0}: Sequence descriptions in the FASTQ file don't match "
                                "({1!r} != {2!r}).\n"
                                "The second sequence description must be either empty "
                                "or equal to the first description.".format(i+1,
                                    name, line[1:]))
                        name2 = name
                    else:
                        name2 = ''
            elif i == 3:
                if len(line) == len(sequence) - strip:
                    qualities = line[:strip]
                else:
                    qualities = line.rstrip('\r\n')
                yield sequence_class(name, sequence, qualities, name2=name2)
            i = (i + 1) % 4
        if i != 0:
            raise FormatError("FASTQ file ended prematurely")



# coding: utf-8
"""This module implements all the read modifications that atropos supports.
A modifier must be callable. It is implemented as a function if no parameters
need to be stored, and as a class with a __call__ method if there are parameters
(or statistics).
"""
from collections import OrderedDict
import copy
import re
from atropos import AtroposError
from atropos.align import (
    Aligner, InsertAligner, SEMIGLOBAL, START_WITHIN_SEQ1, STOP_WITHIN_SEQ2)
from atropos.util import BASE_COMPLEMENTS, reverse_complement, mean, quals2ints
from .qualtrim import quality_trim_index, nextseq_trim_index

# Base classes

class Modifier(object):
    """Base clas for modifiers.
    """
    @property
    def name(self):
        """Modifier name.
        """
        return self.__class__.__name__
    
    @property
    def description(self):
        """Modifier description (for display).
        """
        return getattr(self, 'display_str', self.name)
    
    def summarize(self):
        """Returns a summary of the modifier's activity as a dict.
        """
        return {}

class ReadPairModifier(Modifier):
    """Base class of modifiers that edit a pair of reads simultaneously.
    """
    def __call__(self, read1, read2):
        raise NotImplementedError()

class Trimmer(Modifier):
    """Base class of modifiers that trim bases from reads.
    """
    def __init__(self):
        self.trimmed_bases = 0
    
    def __call__(self, read):
        raise NotImplementedError()
    
    def subseq(self, read, begin=0, end=None):
        """Returns a subsequence of a read.
        
        Args:
            read: The read to trim.
            begin: The first base of the subsequnce.
            end: The last base of the subsequence, or None for len(read).
        """
        if begin or (end is not None):
            front_bases, back_bases, new_read = read.subseq(begin, end)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read
    
    def clip(self, read, front=0, back=0):
        """Returns a read with bases trimmed off the front/back.
        
        Args:
            read: The read to trim.
            front: The number of bases to trim from the front.
            back: The (negative) number of bases to trim from the back.
        """
        if (front or back) and len(read) > 0:
            front_bases, back_bases, new_read = read.clip(front, back)
            self.trimmed_bases += front_bases + back_bases
            return new_read
        else:
            return read
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(bp_trimmed=self.trimmed_bases)

# Modifiers

class AdapterCutter(Modifier):
    """Repeatedly find one of multiple adapters in reads. The number of times
    the search is repeated is specified by the times parameter.
    
    Args:
        adapters: List of Adapter objects.
        times: Number of times to trim.
        action: What to do with a found adapter: None, 'trim', or 'mask'
    """
    def __init__(self, adapters=None, times=1, action='trim'):
        super(AdapterCutter, self).__init__()
        self.adapters = adapters or []
        self.times = times
        self.action = action
        self.with_adapters = 0

    def _best_match(self, read):
        """Find the best matching adapter in the given read.
        
        Returns:
            Either a Match instance or None if there are no matches.
        """
        best = None
        for adapter in self.adapters:
            match = adapter.match_to(read)
            if match is None:
                continue
            
            # the no. of matches determines which adapter fits best
            if best is None or match.matches > best.matches:
                best = match
        return best

    def __call__(self, read):
        """Determine the adapter that best matches the given read.
        Since the best adapter is searched repeatedly, a list
        of Match instances is returned, which
        need to be applied consecutively to the read.
        The list is empty if there are no adapter matches.
        
        The read is converted to uppercase before it is compared to the adapter
        sequences.
        
        Cut found adapters from a single read. Return modified read.
        """
        if len(read) == 0:
            return read
        
        matches = []
        
        # try at most self.times times to remove an adapter
        trimmed_read = read
        for _ in range(self.times):
            match = self._best_match(trimmed_read)
            if match is None:
                # nothing found
                break
            matches.append(match)
            trimmed_read = match.adapter.trimmed(match)
        
        if not matches:
            trimmed_read.match = None
            trimmed_read.match_info = None
            return trimmed_read
        
        if __debug__:
            assert len(trimmed_read) < len(read), \
                "Trimmed read isn't shorter than original"
        
        if self.action == 'trim':
            # read is already trimmed, nothing to do
            pass
        elif self.action == 'mask':
            # add N from last modification
            masked_sequence = trimmed_read.sequence
            for match in sorted(matches, reverse=True, key=lambda m: m.astart):
                nstr = 'N' * (
                    len(match.read.sequence) -
                    len(match.adapter.trimmed(match).sequence))
                # add N depending on match position
                if match.front:
                    masked_sequence = nstr + masked_sequence
                else:
                    masked_sequence += nstr
            # set masked sequence as sequence with original quality
            trimmed_read.sequence = masked_sequence
            trimmed_read.qualities = matches[0].read.qualities
            
            assert len(trimmed_read.sequence) == len(read)
        elif self.action is None:
            trimmed_read = read
        
        trimmed_read.match = matches[-1]
        trimmed_read.match_info = [match.get_info_record() for match in matches]
        
        self.with_adapters += 1
        return trimmed_read
    
    def summarize(self):
        adapters_summary = OrderedDict()
        for adapter in self.adapters:
            adapters_summary[adapter.name] = adapter.summarize()
        return dict(
            records_with_adapters=self.with_adapters,
            adapters=adapters_summary)

# Other error correction approaches:
# https://www.ncbi.nlm.nih.gov/pubmed/25161220

class ErrorCorrectorMixin(object):
    """Provides a method for error correction.
    
    Args:
        mismatch_action: The action to take when a mismatch between the
            overlapping portions of read1 and read2 is encountered. Valid
            values are 'liberal', 'conservative', 'N'.
        min_qual_difference: When mismatch_action=='conservative', the minimum
            difference in base quality between read1 and read2 required to
            perform the correction.
    """
    def __init__(self, mismatch_action=None, min_qual_difference=1):
        self.mismatch_action = mismatch_action
        self.r1r2_min_qual_difference = min_qual_difference
        self.r2r1_min_qual_difference = -1 * min_qual_difference
        self.corrected_pairs = 0
        self.corrected_bp = [0, 0]
    
    def correct_errors(self, read1, read2, insert_match):
        """Correct errors in overlapping reads.
        
        Args:
            read1, read2: The reads.
            insert_match: The match info telling where the reads overlap.
        """
        # Do not attempt to correct an already corrected read
        if read1.corrected > 0 or read2.corrected > 0:
            return
        
        # read2 reverse-complement is the reference, read1 is the query
        r1_seq = list(read1.sequence)
        r2_seq = list(read2.sequence)
        len2 = len(r2_seq)
        
        has_quals = read1.qualities and read2.qualities
        if has_quals:
            r1_qual = list(read1.qualities)
            r2_qual = list(read2.qualities)
        elif self.mismatch_action in ('liberal', 'conservative'):
            raise ValueError(
                "Cannot perform quality-based error correction on reads "
                "lacking quality information")
        
        r1_start = insert_match[2]
        r1_end = insert_match[3]
        r1_changed = 0
        r2_start = len2 - insert_match[1]
        r2_end = len2 - insert_match[0]
        r2_changed = 0
        quals_equal = []
        
        for i, j in zip(
                range(r1_start, r1_end), range(r2_end - 1, r2_start, -1)):
            base1 = r1_seq[i]
            base2 = BASE_COMPLEMENTS[r2_seq[j]]
            if base1 == base2:
                continue
            if self.mismatch_action == 'N':
                r1_seq[i] = 'N'
                r2_seq[j] = 'N'
                r1_changed += 1
                r2_changed += 1
            elif base1 == 'N':
                r1_seq[i] = base2
                if has_quals:
                    r1_qual[i] = r2_qual[j]
                r1_changed += 1
            elif base2 == 'N':
                r2_seq[j] = BASE_COMPLEMENTS[base1]
                if has_quals:
                    r2_qual[j] = r1_qual[i]
                r2_changed += 1
            elif has_quals:
                diff = ord(r1_qual[i]) - ord(r2_qual[j])
                if diff >= self.r1r2_min_qual_difference:
                    r2_seq[j] = BASE_COMPLEMENTS[base1]
                    r2_qual[j] = r1_qual[i]
                    r2_changed += 1
                elif diff <= self.r2r1_min_qual_difference:
                    r1_seq[i] = base2
                    r1_qual[i] = r2_qual[j]
                    r1_changed += 1
                elif self.mismatch_action == 'liberal':
                    quals_equal.append((i, j, base1, base2))
        
        if quals_equal:
            mean_qual1 = mean([ord(b) for b in r1_qual[r1_start:r1_end]])
            mean_qual2 = mean([ord(b) for b in r2_qual[r2_start:r2_end]])
            # Only make the corrections if one read is significantly better
            # than the other.
            # TODO: this method of determining whether one read is better
            # than the other is crude - come up with something better.
            diff = mean_qual1 - mean_qual2
            if diff > 1:
                # read1 is better than read2
                for i, j, base1, base2 in quals_equal:
                    r2_seq[j] = BASE_COMPLEMENTS[base1]
                    r2_qual[j] = r1_qual[i]
                    r2_changed += 1
            elif diff < -1:
                # read2 is better than read1
                for i, j, base1, base2 in quals_equal:
                    r1_seq[i] = base2
                    r1_qual[i] = r2_qual[j]
                    r1_changed += 1
        
        if r1_changed or r2_changed:
            self.corrected_pairs += 1
            if r1_changed:
                self.corrected_bp[0] += r1_changed
                read1.sequence = ''.join(r1_seq)
                read1.corrected = r1_changed
                if has_quals:
                    read1.qualities = ''.join(r1_qual)
            if r2_changed:
                self.corrected_bp[1] += r2_changed
                read2.sequence = ''.join(r2_seq)
                read2.corrected = r2_changed
                if has_quals:
                    read2.qualities = ''.join(r2_qual)
    
    def summarize(self):
        """Returns a summary dict.
        """
        return dict(
            records_corrected=self.corrected_pairs,
            bp_corrected=self.corrected_bp)

class InsertAdapterCutter(ReadPairModifier, ErrorCorrectorMixin):
    """AdapterCutter that uses InsertAligner to first try to identify
    insert overlap before falling back to semi-global adapter alignment.
    
    Args:
        adapter1, adapter2: Adapters.
        action: Action to take on adapter match: trim, mask (replace adapter
            with N's), lower (convert adapter bases to lower case),
            or None.
        mismatch_action: How to deal with mismatches. See
            :class:`ErrorCorrectorMixin`.
        symmetric: Whether to assume that the adapter should appear in the
            same place on overlapping reads.
        min_insert_overlap: Minimum overlap required between reads to be
            considered an insert match.
        aligner_args: Additional arguments to :class:`InsertAligner`.
    """
    def __init__(
            self, adapter1, adapter2, action='trim', mismatch_action=None,
            symmetric=True, min_insert_overlap=1, **aligner_args):
        ErrorCorrectorMixin.__init__(self, mismatch_action)
        self.adapter1 = adapter1
        self.adapter2 = adapter2
        self.aligner = InsertAligner(
            adapter1.sequence, adapter2.sequence,
            min_insert_overlap=min_insert_overlap, **aligner_args)
        self.min_insert_len = min_insert_overlap
        self.action = action
        self.symmetric = symmetric
        self.with_adapters = [0, 0]
    
    def __call__(self, read1, read2):
        read_lengths = [len(r) for r in (read1, read2)]
        if any(l < self.min_insert_len for l in read_lengths):
            return (read1, read2)
        
        match = self.aligner.match_insert(read1.sequence, read2.sequence)
        read1.insert_overlap = read2.insert_overlap = (match is not None)
        insert_match = None
        correct_errors = False
        
        if match:
            insert_match, adapter_match1, adapter_match2 = match
            correct_errors = self.mismatch_action and insert_match[5] > 0
        else:
            adapter_match1 = self.adapter1.match_to(read1)
            adapter_match2 = self.adapter2.match_to(read2)
            # If the adapter matches are complementary, perform error correction
            if (
                    self.mismatch_action and adapter_match1 and
                    adapter_match2 and
                    adapter_match1.rstart == adapter_match2.rstart):
                insert_match = (
                    read_lengths[1] - adapter_match1.rstart,
                    read_lengths[1], 0, adapter_match1.rstart)
                correct_errors = True
        
        # If exactly one of the two alignments failed and symmetrix is True,
        # duplicate the good alignment
        if self.symmetric and sum(
                bool(m) for m in (adapter_match1, adapter_match2)) == 1:
            if adapter_match1:
                adapter_match2 = adapter_match1.copy()
            else:
                adapter_match1 = adapter_match2.copy()
            if self.mismatch_action and not insert_match:
                # Assume that the symmetric read segments overlap and
                # perform error correction
                insert_match = (
                    read_lengths[1] - adapter_match1.rstart,
                    read_lengths[1], 0, adapter_match1.rstart)
                correct_errors = True
        
        if correct_errors:
            self.correct_errors(read1, read2, insert_match)
        
        return (
            self.trim(read1, self.adapter1, adapter_match1, 0),
            self.trim(read2, self.adapter2, adapter_match2, 1))
    
    def trim(self, read, adapter, match, read_idx):
        """Trim an adapter from a read.
        
        Args:
            read: The read to trim from.
            adapter: The Adapter to trim.
            match: The match details.
            read_idx: 0/1
        """
        if not match:
            read.match = None
            read.match_info = None
            return read
        
        match.adapter = adapter
        match.read = read
        match.front = False
        
        if self.action is None or match.rstart >= len(read):
            trimmed_read = read
        
        else:
            trimmed_read = adapter.trimmed(match)
            
            if self.action == 'mask':
                # add N from last modification
                masked_sequence = trimmed_read.sequence
                masked_sequence += 'N' * (len(read) - len(trimmed_read))
                # set masked sequence as sequence with original quality
                trimmed_read.sequence = masked_sequence
                trimmed_read.qualities = read.qualities
            elif self.action == 'lower':
                # TODO: offer option to mask with lower-case of trimmed base
                # This will happen as part of the refactoring to modify
                # Sequences in-place.
                pass
        
        trimmed_read.match = match
        trimmed_read.match_info = [match.get_info_record()]
        
        self.with_adapters[read_idx] += 1
        return trimmed_read
    
    def summarize(self):
        """Returns a summary dict.
        """
        adapters_summary = tuple(
            { adapter.name : adapter.summarize() }
            for adapter in (self.adapter1, self.adapter2))
        summary = dict(
            records_with_adapters=self.with_adapters,
            adapters=adapters_summary)
        if self.mismatch_action:
            summary.update(ErrorCorrectorMixin.summarize(self))
        return summary

class OverwriteRead(ReadPairModifier):
    """If one read is of significantly worse quality than the other, overwrite
    the poor quality read with the good-quality read.
    
    Computes the mean quality over the first `window_size` bases. If one read
    has mean quality < `worse_read_min_quality` and the other read has mean
    quality >= `better_read_min_quality`, replace the entire worse read with
    the better read.
    
    Args:
        worse_read_min_quality: The minimum quality below which a read is
            considered 'poor'
        better_read_min_quality: The quality above which a read is considered
            'good'
        window_size: The number of bases over which to compute mean quality
        base: The base for ascii-to-int quality conversion
        summary_fn: Function to summarize base quality (default=mean)
    """
    def __init__(
            self, worse_read_min_quality, better_read_min_quality,
            window_size, base=33, summary_fn=mean):
        self.worse_read_min_quality = worse_read_min_quality
        self.better_read_min_quality = better_read_min_quality
        self.window_size = window_size
        self.base = base
        self.summary_fn = summary_fn
    
    def __call__(self, read1, read2):
        if len(read1) < self.window_size or len(read2) < self.window_size:
            return (read1, read2)
        if not (read1.qualities and read2.qualities):
            raise ValueError(
                "OverwriteRead modifier does not work with reads "
                "lacking base qualities.")
        qual1 = list(quals2ints(read1.qualities[:self.window_size], self.base))
        summ1 = self.summary_fn(qual1)
        
        qual2 = list(quals2ints(read2.qualities[:self.window_size], self.base))
        summ2 = self.summary_fn(qual2)
        
        if (
                summ1 < self.worse_read_min_quality and
                summ2 >= self.better_read_min_quality):
            # TODO: not sure what the right value is here
            read2.corrected = 1
            read1 = read2.reverse_complement()
        elif (
                summ2 < self.worse_read_min_quality and
                summ1 >= self.better_read_min_quality):
            read1.corrected = 1
            read2 = read1.reverse_complement()
        
        return (read1, read2)

class UnconditionalCutter(Trimmer):
    """A modifier that unconditionally removes the first n or the last n bases
    from a read. Equivalent to MinCutter with count_trimmed=False and
    only_trimmed=False, but UnconditionalCutter is applied before adapter
    trimming.
    
    If the length is positive, the bases are removed from the beginning of the
    read. If the length is negative, the bases are removed from the end of the
    read.
    """
    display_str = "Cut unconditionally"
    
    def __init__(self, lengths=None):
        super().__init__()
        self.front_length = self.back_length = 0
        if lengths:
            self.front_length = sum(front for front in lengths if front > 0)
            self.back_length = sum(back for back in lengths if back < 0)
    
    def __call__(self, read):
        return self.clip(read, self.front_length, self.back_length)

class MinCutter(Trimmer):
    """Ensure that a minimum number of bases have been trimmed off each end.
    
    Args:
        lengths: Sequence of bases to trim. Numbers > 0 are summed to the
            total bases trimmed from the front of the read, and numbers < 0
            are summed to the total bases trimmed from the end of the read.
        count_trimmed: Whether to consider bases cut before or during adapter
            trimming when counting the number of bases that have already been
            cut.
        only_trimmed: Only cut read ends if they have already been
            adapter-trimmed.
    """
    display_str = "Cut conditionally"
    
    def __init__(self, lengths=None, count_trimmed=True, only_trimmed=False):
        super().__init__()
        self.front_length = self.back_length = 0
        if lengths:
            self.front_length = sum(front for front in lengths if front > 0)
            self.back_length = sum(back for back in lengths if back < 0)
        self.count_trimmed = count_trimmed
        self.only_trimmed = only_trimmed
    
    def __call__(self, read):
        trim_front = trim_back = True
        if self.only_trimmed:
            if read.match:
                is_front = [match.is_front for match in read.match_info]
                if not any(is_front):
                    trim_front = False
                elif all(is_front):
                    trim_back = False
            else:
                return read
        
        # TODO: distinguish between adapter trimming and other trimming.
        # For things like Methyl-Seq, we want to trim additional bases
        # after adapter trimming, but we want to count other post-adapter
        # trimmed bases toward the minimum length
        def to_trim(offset, is_front):
            """Returns number of bases that need to be trimmed.
            """
            if self.count_trimmed:
                trimmed = read.clipped[offset] + read.clipped[offset+2]
                if read.match:
                    trimmed += sum(
                        i.rsize_total
                        for i in read.match_info
                        if is_front == i.is_front)
            elif read.match:
                trimmed = read.clipped[offset+2]
            else:
                trimmed = read.clipped[offset]
            
            if is_front:
                return max(self.front_length - trimmed, 0)
            else:
                return min(trimmed + self.back_length, 0)
        
        return self.clip(
            read,
            to_trim(0, True) if trim_front else 0,
            to_trim(1, False) if trim_back else 0)

class LengthTagModifier(Modifier):
    """Replace "length=..." strings in read names.
    """
    def __init__(self, length_tag="length="):
        self.regex = re.compile(r"\b" + length_tag + r"[0-9]*\b")
        self.length_tag = length_tag
    
    def __call__(self, read):
        read = read[:]
        if read.name.find(self.length_tag) >= 0:
            read.name = self.regex.sub(
                self.length_tag + str(len(read.sequence)), read.name)
        return read

class SuffixRemover(Modifier):
    """Remove a given suffix from read names.
    """
    def __init__(self, suffixes=None):
        self.suffixes = suffixes or []
    
    def __call__(self, read):
        name = read.name
        for suffix in self.suffixes:
            if name.endswith(suffix):
                name = name[:-len(suffix)]
        read = read[:]
        read.name = name
        return read

class PrefixSuffixAdder(Modifier):
    """Add a suffix and a prefix to read names.
    """
    def __init__(self, prefix="", suffix=""):
        self.prefix = prefix
        self.suffix = suffix
    
    def __call__(self, read):
        read = read[:]
        adapter_name = 'no_adapter'
        if read.match is not None:
            adapter_name = read.match.adapter.name
        read.name = self.prefix.replace('{name}', adapter_name) + read.name + \
            self.suffix.replace('{name}', adapter_name)
        return read

class DoubleEncoder(Modifier):
    """Double-encode colorspace reads, using characters ACGTN to represent
    colors.
    """
    def __init__(self):
        self.double_encode_trans = str.maketrans('0123.', 'ACGTN')
    
    def __call__(self, read):
        read = read[:]
        read.sequence = read.sequence.translate(self.double_encode_trans)
        return read

class ZeroCapper(Modifier):
    """Change negative quality values of a read to zero
    """
    def __init__(self, quality_base=33):
        qbase = quality_base
        self.zero_cap_trans = str.maketrans(
            ''.join(map(chr, range(qbase))), chr(qbase) * qbase)
    
    def __call__(self, read):
        read = read[:]
        read.qualities = read.qualities.translate(self.zero_cap_trans)
        return read

class PrimerTrimmer(Trimmer):
    """Trims primer base from colorspace reads.
    """
    display_str = "Primer-trimmed"
    
    def __call__(self, read):
        read = self.clip(read, 1)
        read.primer = ''
        return read

class NextseqQualityTrimmer(Trimmer):
    """NextSeq-specific quality trimmer.
    """
    display_str = "Quality trimmed (NextSeq)"
    
    def __init__(self, cutoff=0, base=33):
        super(NextseqQualityTrimmer, self).__init__()
        self.cutoff = cutoff
        self.base = base
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        stop = nextseq_trim_index(read, self.cutoff, self.base)
        return self.subseq(read, end=stop)

class QualityTrimmer(Trimmer):
    """Trim bases from the start/end of reads based on their qualities.
    """
    display_str = "Quality-trimmed"
    
    def __init__(self, cutoff_front=0, cutoff_back=0, base=33):
        super(QualityTrimmer, self).__init__()
        self.cutoff_front = cutoff_front
        self.cutoff_back = cutoff_back
        self.base = base
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        start, stop = quality_trim_index(
            read.qualities, self.cutoff_front, self.cutoff_back, self.base)
        return self.subseq(read, start, stop)

class NEndTrimmer(Trimmer):
    """Trims Ns from the 3' and 5' end of reads.
    """
    display_str = "End Ns trimmed"
    
    def __init__(self):
        super(NEndTrimmer, self).__init__()
        self.start_trim = re.compile(r'^N+')
        self.end_trim = re.compile(r'N+$')
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        sequence = read.sequence
        start_cut = self.start_trim.match(sequence)
        end_cut = self.end_trim.search(sequence)
        start_cut = start_cut.end() if start_cut else 0
        end_cut = end_cut.start() if end_cut else len(read)
        return self.subseq(read, start_cut, end_cut)

class RRBSTrimmer(MinCutter):
    """Sequences that are adapter-trimmed are further trimmed 2 bp on the 3'
    end to remove potential methylation-biased bases from the end-repair
    reaction.
    """
    display_str = "RRBS-trimmed"
    
    def __init__(self, trim_5p=0, trim_3p=2):
        super().__init__(
            (trim_5p, -1 * trim_3p), count_trimmed=False, only_trimmed=True)

class NonDirectionalBisulfiteTrimmer(Modifier):
    """For non-directional RRBS/WGBS libraries (which implies that they were
    digested using MspI), sequences that start with either 'CAA' or 'CGA' will
    have 2 bp trimmed off the 5' end to remove potential methylation-biased
    bases from the end-repair reaction. Additionally, for RRBS reads, if CAA/CGA
    is not trimmed *and* the read has been adapter-trimmed, a minimum number of
    bases is trimmed from the 3' end.
    """
    display_str = "Bisulfite-trimmed (Non-directional)"
    _regex = re.compile(r"^C[AG]A")
    
    def __init__(self, trim_5p=2, trim_3p=2, rrbs=False):
        self._non_directional_cutter = MinCutter(
            [trim_5p], count_trimmed=False, only_trimmed=False)
        self.rrbs = rrbs
        if rrbs:
            self._rrbs_cutter = RRBSTrimmer(trim_3p)
    
    def __call__(self, read):
        if len(read) == 0:
            return read
        cutter = None
        if self._regex.match(read.sequence):
            cutter = self._non_directional_cutter
        elif self.rrbs:
            cutter = self._rrbs_cutter
        return cutter(read) if cutter else read
    
    def summarize(self):
        """Returns the number of trimmed bases.
        """
        bp_trimmed = (
            self._rrbs_cutter.trimmed_bases +
            self._non_directional_cutter.trimmed_bases)
        return dict(bp_trimmed=bp_trimmed)

class TruSeqBisulfiteTrimmer(MinCutter):
    """EpiGnome reads are trimmed at least 6 bp on the 5' end.
    """
    display_str = "Bisulfite-trimmed (EpiGnome/TruSeq)"
    
    def __init__(self):
        super().__init__((6,), count_trimmed=True, only_trimmed=False)

class SwiftBisulfiteTrimmer(ReadPairModifier):
    """For WGBS libraries prepared with the Swift Accel-NGS kit, 10 bp are
    trimmed  off the end of read1 and the beginning of read2.
    """
    display_str = "Bisulfite-trimmed (Swift)"
    
    def __init__(self, trim_5p1=0, trim_3p1=10, trim_5p2=10, trim_3p2=0):
        self._read1_cutter = MinCutter(
            (trim_5p1, -1 * trim_3p1), count_trimmed=False, only_trimmed=False)
        self._read2_cutter = MinCutter(
            (trim_5p2, -1 * trim_3p2), count_trimmed=False, only_trimmed=False)
    
    def __call__(self, read1, read2):
        return (self._read1_cutter(read1), self._read2_cutter(read2))
    
    def summarize(self):
        return dict(bp_trimmed=(
            self._read1_cutter.trimmed_bases,
            self._read2_cutter.trimmed_bases))

# TODO: InsertAdapterCutter should save the insert match, and
# MergeOverlapping should use that rather than doing another alignment

class MergeOverlapping(ReadPairModifier, ErrorCorrectorMixin):
    """Merge overlaping reads. The merged reads are stored in read1.
    """
    def __init__(self, min_overlap=0.9, error_rate=0.1, mismatch_action=None):
        ErrorCorrectorMixin.__init__(self, mismatch_action)
        self.min_overlap = int(min_overlap) if min_overlap > 1 else min_overlap
        self.error_rate = error_rate
    
    def __call__(self, read1, read2):
        len1 = len(read1.sequence)
        len2 = len(read2.sequence)
        min_overlap = self.min_overlap
        if min_overlap <= 1:
            min_overlap = max(2, round(self.min_overlap * min(len1, len2)))
        
        if len1 < min_overlap or len2 < min_overlap:
            return (read1, read2)
        
        insert_matched = read1.insert_overlap and read2.insert_overlap
        
        if insert_matched:
            # If we've already determined that there is an insert overlap
            # with a 3' overhang, we can constrain our alignment
            aflags = START_WITHIN_SEQ1 | STOP_WITHIN_SEQ2
        else:
            aflags = SEMIGLOBAL
        # align read1 to read2 reverse-complement to be compatible with
        # InsertAligner
        read2_rc = reverse_complement(read2.sequence)
        aligner = Aligner(read2_rc, self.error_rate, aflags)
        alignment = aligner.locate(read1.sequence)
        
        if alignment:
            r2_start, r2_stop, r1_start, r1_stop, matches, errors = alignment
            if matches >= min_overlap:
                # Only correct errors if we haven't already done correction in
                # the InsertAligner
                if (
                        self.mismatch_action and errors > 0 and
                        not insert_matched and
                        read1.corrected == 0 and read2.corrected == 0):
                    self.correct_errors(read1, read2, alignment)
                
                if r2_start == 0 and r2_stop == len2:
                    # r2 is fully contained in r1
                    pass
                elif r1_start == 0 and r1_stop == len1:
                    # r1 is fully contained in r2
                    read1.sequence = read2_rc
                    read1.qualities = "".join(reversed(read2.qualities))
                elif r1_start > 0:
                    read1.sequence += read2_rc[r2_stop:]
                    if read1.qualities and read2.qualities:
                        read1.qualities += "".join(
                            reversed(read2.qualities))[r2_stop:]
                elif r2_start > 0:
                    read1.sequence = read2_rc + read1.sequence[r1_stop:]
                    if read1.qualities and read2.qualities:
                        read1.qualities = (
                            "".join(reversed(read2.qualities)) +
                            read1.qualities[r1_stop:])
                else:
                    raise AtroposError(
                        "Invalid alignment while trying to merge read "
                        "{}: {}".format(
                            read1.name, ",".join(str(i) for i in alignment)))
                
                read1.merged = True
                read2 = None
                
        return (read1, read2)

class Modifiers(object):
    """Base for classes that manage multiple modifiers.
    """
    def __init__(self):
        self.modifiers = []
        self.modifier_indexes = {}
    
    def add_modifier(self, mod_class, read=1|2, **kwargs):
        """Add a modifier of the specified type for one or both reads.
        
        Args:
            mod_class: The type of modifier to add.
            read: Which reads this modifier should modify; 1, 2, or 3 (1|2).
            kwargs: Additional keyword arguments to the modifier constructor.
        """
        raise NotImplementedError()
    
    def add_modifier_pair(self, mod_class, read1_args=None, read2_args=None):
        """Add a modifier for both reads in a pair.
        
        Args:
            mod_class: The type of modifier to add. Cannot be a subclass of
                ReadPairModifier.
            read1_args, read2_args: Additional keyword arguments to pass to the
                modifier constructors.
        """
        raise NotImplementedError()
    
    def _add_modifiers(self, mod_class, mods):
        idx = len(self.modifiers)
        self.modifiers.append(mods)
        if mod_class in self.modifier_indexes:
            self.modifier_indexes[mod_class].append(idx)
        else:
            self.modifier_indexes[mod_class] = [idx]
        return idx
    
    def has_modifier(self, mod_class):
        """Returns True if a modifier of the specified type has been added.
        """
        return mod_class in self.modifier_indexes
    
    def get_modifiers(self, mod_class=None, read=None):
        """Returns a list of modifiers that have been added.
        
        Args:
            mod_class: Restrict modifiers to those of a certain type.
            read: Return only the modifiers for a given read (1 or 2)
        """
        if mod_class is None:
            mods = copy.copy(self.modifiers)
        elif mod_class in self.modifier_indexes:
            mods = [self.modifiers[i] for i in self.modifier_indexes[mod_class]]
        else:
            mods = []
        
        if not (mods and read):
            return mods
        
        read_mods = []
        for mod in mods:
            if isinstance(mod, ReadPairModifier):
                read_mods.append(mod)
            elif mod[read-1] is not None:
                read_mods.append(mod[read-1])
        return read_mods
    
    def get_adapters(self):
        """Returns the adapters from the AdapterCutter or InsertAdapterCutter
        modifier, if any.
        
        Returns:
            A list [[read1_adapters], [read2_adapters]].
        """
        adapters = [[], []]
        if self.has_modifier(AdapterCutter):
            mod1, mod2 = self.get_modifiers(AdapterCutter)[0]
            if mod1:
                adapters[0] = mod1.adapters
            if mod2:
                adapters[1] = mod2.adapters
        elif self.has_modifier(InsertAdapterCutter):
            mod = self.get_modifiers(InsertAdapterCutter)[0]
            adapters[0] = [mod.adapter1]
            adapters[1] = [mod.adapter2]
        return adapters
    
    def modify(self, read1, read2=None):
        """Apply registered modifiers to a read/pair.
        
        Args:
            read1, read2: The reads to modify.
        
        Returns:
            A tuple of modified reads (read1, read2).
        """
        raise NotImplementedError()
    
    def summarize(self):
        """Returns a summary dict.
        """
        raise NotImplementedError()
    
class SingleEndModifiers(Modifiers):
    """Manages modifiers for single-end data.
    """
    def add_modifier(self, mod_class, read=1, **kwargs):
        if read != 1:
            raise ValueError("'read' must be 1 for single-end data")
        return self._add_modifiers(mod_class, [mod_class(**kwargs), None])
    
    def add_modifier_pair(self, mod_class, read1_args=None, read2_args=None):
        if read1_args is not None:
            return self.add_modifier(mod_class, **read1_args)
    
    def modify(self, read1, read2=None):
        for mods in self.modifiers:
            read1 = mods[0](read1)
        return (read1,)
    
    def summarize(self):
        summary = {}
        for mods in self.modifiers:
            mod = mods[0]
            summary[mod.name] = dict(
                (key, (value,))
                for key, value in mod.summarize().items())
            summary[mod.name]['desc'] = mod.description
        return summary

class PairedEndModifiers(Modifiers):
    """Manages modifiers for paired-end data.
    """
    def __init__(self, paired):
        super().__init__()
        self.paired = paired
    
    def add_modifier(self, mod_class, read=1|2, **kwargs):
        if issubclass(mod_class, ReadPairModifier):
            if self.paired != "both" and read == 1|2:
                raise ValueError(
                    "Must have paired-end reads to use modifer {}".format(
                        mod_class))
            mods = mod_class(**kwargs)
        else:
            mods = [None, None]
            if read & 1 > 0:
                mods[0] = mod_class(**kwargs)
            if read & 2 > 0 and self.paired == "both":
                mods[1] = mod_class(**kwargs)
            if not any(mods):
                return None
        return self._add_modifiers(mod_class, mods)
    
    def add_modifier_pair(self, mod_class, read1_args=None, read2_args=None):
        mods = [None, None]
        if read1_args is not None:
            mods[0] = mod_class(**read1_args)
        if read2_args is not None and self.paired == "both":
            mods[1] = mod_class(**read2_args)
        if any(mods):
            return self._add_modifiers(mod_class, mods)
    
    def modify(self, read1, read2=None):
        for mods in self.modifiers:
            if isinstance(mods, ReadPairModifier):
                read1, read2 = mods(read1, read2)
            else:
                if mods[0] is not None:
                    read1 = mods[0](read1)
                if mods[1] is not None:
                    read2 = mods[1](read2)
        return (read1, read2)
    
    def summarize(self):
        summary = {}
        for mods in self.modifiers:
            if isinstance(mods, ReadPairModifier):
                summary[mods.name] = mods.summarize()
                summary[mods.name]['desc'] = mods.description
            elif any(mods):
                name = desc = keys = None
                summ1 = summ2 = {}
                if mods[0]:
                    name = mods[0].name
                    desc = mods[0].description
                    summ1 = mods[0].summarize()
                    if summ1:
                        keys = summ1.keys()
                if mods[1]:
                    summ2 = mods[1].summarize()
                    if summ2:
                        if name:
                            assert name == mods[1].name
                            assert desc == mods[1].description
                            assert set(keys) == set(summ2.keys())
                        else:
                            name = mods[1].name
                            desc = mods[1].description
                            keys = summ2.keys()
                if keys:
                    summary[name] = dict(
                        (key, (summ1.get(key, None), summ2.get(key, None)))
                        for key in keys)
                    summary[name]['desc'] = desc
        return summary



"""Estimate the empircal error rate.
"""
from collections import defaultdict
import csv
import re
from atropos import AtroposError
from atropos.commands.base import BaseCommandRunner
from atropos.io import open_output
from atropos.util import qual2prob

class CommandRunner(BaseCommandRunner):
    name = 'error'
    
    def __call__(self):
        if not self.delivers_qualities:
            raise ValueError(
                "Cannot estimate error rate without base qualities")
        
        if self.algorithm == 'quality':
            estimator_class = BaseQualityErrorEstimator
        elif self.algorithm == 'shadow':
            estimator_class = ShadowRegressionErrorEstimator
        
        if self.paired:
            estimator = PairedErrorEstimator(
                max_read_len=self.max_bases,
                estimator_class=estimator_class)
        else:
            estimator = estimator_class(max_read_len=self.max_bases)
        
        self.summary['error_estimator'] = estimator
        
        estimator.consume_all_batches(self)
        
        return 0

class ErrorEstimator(object):
    """Base class for error estimators.
    """
    def consume_all_batches(self, batch_iterator):
        """Consume all batches in `batch_iterator`.
        """
        for batch_meta, batch in batch_iterator:
            if batch_meta['size'] == 0:
                continue
            for read in batch:
                self.consume(read)
    
    def consume(self, read):
        """Consume a sequence record.
        """
        raise NotImplementedError()
    
    def estimate(self):
        """Returns an estimate of the error rate.
        """
        raise NotImplementedError()
    
    # TODO: move report to the report package, have summarize update the
    # summary dict.
    
    def summarize(self, outstream, name=None, show_details=True, indent="  "):
        """Print a summary report.
        """
        print("", file=outstream)
        if name is not None:
            header = "File: {}".format(name)
            print(header, file=outstream)
            print('-' * len(header), file=outstream)
        overall, details = self.estimate()
        print("Error rate: {:.2%}".format(overall), file=outstream)
        if details and show_details:
            print("Details:\n", file=outstream)
            self.print_details(details, outstream, indent)
    
    def print_details(self, details, outstream, indent):
        """Print additional details.
        """
        pass

class BaseQualityErrorEstimator(ErrorEstimator):
    """Simple error estimation using base qualities. It is well-known that base
    qualities significantly overestimate true error rates, so take the estimate
    with a grain of salt.
    """
    def __init__(self, max_read_len=None):
        self.total_qual = 0.0
        self.total_len = 0
        self.max_read_len = max_read_len
    
    def consume(self, read):
        quals = read.qualities
        readlen = len(quals)
        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            quals = quals[:readlen]
        self.total_qual += sum(qual2prob(qchar) for qchar in quals)
        self.total_len += readlen
    
    def estimate(self):
        return (self.total_qual / self.total_len, None)

# Error estimation using shadow counts

FILTER_RE = re.compile("A+|C+|G+|T+|.*N.*")

SHADOW_REGRESSION_R_SCRIPT = """
library(ShadowRegression)
errorRates = getErrorRates("{reads}", type="{method}")
write.table(errorRates$perReadER, "{per_read}", sep="\t", quote=F, col.names=F,
            row.names=T)
write.table(errorRates$cycleER, "{per_cycle}", sep="\t", quote=F, col.names=F,
            row.names=T)
"""

class ShadowRegressionErrorEstimator(ErrorEstimator):
    """Re-implementation of the shadow regression method described in:
    Wang et al., "Estimation of sequencing error rates in short reads",
        BMC Bioinformatics 2012 13:185, DOI: 10.1186/1471-2105-13-185
    
    Args:
        method: The differences that are considered in the error rate
            calculation; sub = substitutions, indel = insertions and
            deletions; all = both substitutions and indels.
        max_read_len: The maximum number of bases (starting from the 5' end)
            to consider from each read.
    """
    def __init__(self, method='sub', max_read_len=None, rscript_exe="Rscript"):
        self.seqs = defaultdict(lambda: 0)
        self.total_len = 0
        self.method = method
        self.max_read_len = max_read_len
        self.rscript_exe = rscript_exe
    
    def consume(self, read):
        seq = read.sequence
        readlen = len(seq)
        if self.max_read_len and self.max_read_len < readlen:
            readlen = self.max_read_len
            seq = seq[:readlen]
        if FILTER_RE.fullmatch(seq):
            return
        self.seqs[seq] += 1
        self.total_len += readlen
    
    def _write_read_counts(self, fileobj):
        writer = csv.writer(fileobj, delimiter=" ")
        writer.writerows(sorted(
            self.seqs.items(), reverse=True, key=lambda i: i[1]))
    
    def estimate(self):
        # This is a temporary solution that requires R and the
        # ShadowRegression package. Eventually, this will be
        # replaced with a pure-python implementation.
        import os
        import subprocess
        import tempfile
        
        tempfiles = (tempfile.mkstemp()[1] for i in range(4))
        read_counts, per_read, per_cycle, script_file = tempfiles
        try:
            # write counts to a file
            with open(read_counts, 'wt') as out:
                self._write_read_counts(out)
            
            # execute R script
            script = SHADOW_REGRESSION_R_SCRIPT.format(
                reads=read_counts, method=self.method, per_read=per_read,
                per_cycle=per_cycle)
            with open(script_file, 'wt') as out:
                out.write(script)
            proc = subprocess.Popen(
                [self.rscript_exe, "--vanilla", script_file],
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            with proc:
                stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    raise AtroposError(
                        "R script failed: rc={}; stdout={}; stderr={}".format(
                            proc.returncode, stdout, stderr))
            
            # read the results
            with open(per_read, 'rt') as i:
                reader = csv.reader(i, delimiter="\t")
                per_read_error = dict(reader)
                if len(per_read_error) != 4:
                    raise AtroposError("Invalid output from R script")
            with open(per_cycle, 'rt') as i:
                reader = csv.reader(i, delimiter="\t")
                per_cycle_error = list(row[0:3] for row in reader)
                if not per_cycle_error:
                    raise AtroposError("Invalid output from R script")
            
            return (
                per_read_error["error rate"],
                dict(per_read=per_read_error, per_cycle=per_cycle_error))
        finally:
            for path in tempfiles:
                os.remove(path)
    
    def print_details(self, outstream, details, indent):
        per_read = details['per_read']
        per_cycle = details['per_cycle']
        
        print(
            "{}StdErr: {:.2%}".format(indent, per_read['standard error']),
            file=outstream)
        print("{}Per-cycle rates:".format(indent), file=outstream)
        for cycle in per_cycle:
            print("{}Cycle: {}, Error: {:.2%}, StdErr: {:.2%}".format(
                indent*2, *cycle))

class PairedErrorEstimator(object):
    """Estimator for a pair of input files.
    """
    def __init__(self, max_read_len=None,
                 estimator_class=BaseQualityErrorEstimator):
        self.estimator1 = estimator_class(max_read_len=max_read_len)
        self.estimator2 = estimator_class(max_read_len=max_read_len)
    
    def consume_all_batches(self, batch_iterator):
        """Consume all batches in iterator.
        """
        for batch_meta, batch in batch_iterator:
            if batch_meta['size'] == 0:
                continue
            for read1, read2 in batch:
                self.estimator1.consume(read1)
                self.estimator2.consume(read2)
    
    def estimate(self):
        """Estimate error rates.
        
        Returns:
            Tuple (read1_estimate, read2_estimate).
        """
        return (self.estimator1.estimate(), self.estimator2.estimate())
    
    def summarize(self, outstream, names=None, show_details=True, indent="  "):
        """Print a summary of the error detection.
        """
        print("", file=outstream)
        if names:
            header = "File 1: {}".format(names[0])
            print(header, file=outstream)
            print('-' * len(header), file=outstream)
        err1, details1 = self.estimator1.estimate()
        print("Error rate: {:.2%}".format(err1), file=outstream)
        if show_details and details1:
            print("Details:\n", file=outstream)
            self.estimator1.print_details(details1, outstream, indent)
        print("", file=outstream)
        
        if names:
            header = "File 2: {}".format(names[1])
            print(header, file=outstream)
            print('-' * len(header), file=outstream)
        err2, details2 = self.estimator2.estimate()
        print("Error rate: {:.2%}".format(err2), file=outstream)
        if show_details and details2:
            print("Details:\n", file=outstream)
            self.estimator2.print_details(details2, outstream, indent)
        print("", file=outstream)
        
        len1 = self.estimator1.total_len
        len2 = self.estimator2.total_len
        overall_err = ((err1 * len1) + (err2 * len2)) / (len1+len2)
        print("Overall error rate: {:.2%}".format(overall_err), file=outstream)



"""Report generator for the error command.

TODO: move reporting functionality out of the ErrorEstimator class.
"""
from atropos.commands.reports import BaseReportGenerator
from atropos.io import open_output

class ReportGenerator(BaseReportGenerator):
    def generate_text_report(self, fmt, summary, outfile, **kwargs):
        if fmt == 'txt':
            estimator = summary['error_estimator']
            with open_output(outfile) as out:
                names = summary['input']['input_names']
                if names and len(names) == 1:
                    names = names[0]
                estimator.summarize(out, )
        else:
            super().generate_from_template(fmt, summary, outfile, **kwargs)



# coding: utf-8
"""
Adapters
"""
import itertools
import logging
import os
import pickle
import re
import sys
from urllib.error import URLError
from urllib.request import urlopen
from atropos import align
from atropos.align import Match
from atropos.io.seqio import ColorspaceSequence, FastaReader
from atropos.util import (
    IUPAC_BASES, GC_BASES, MergingDict, NestedDict, CountingDict, Const, 
    reverse_complement)
from atropos.util import colorspace as cs

class AdapterType(object):
    """Struct for adapter type information.
    
    Args:
        name: Adapter type name.
        desc: Adapter type description.
        flags: Alignment flags.
    """
    def __init__(self, name, desc, *flags):
        self.name = name
        self.desc = desc
        self.flags = flags[0]
        for i in range(1, len(flags)):
            self.flags |= flags[i]
    
    def asdict(self):
        """Returns AdapterType fields in a dict.
        """
        return dict(name=self.name, desc=self.desc, flags=Const(self.flags))

ADAPTER_TYPES = dict(
    back=AdapterType(
        'back', "regular 3'",
        align.START_WITHIN_SEQ2,
        align.STOP_WITHIN_SEQ2,
        align.STOP_WITHIN_SEQ1),
    front=AdapterType(
        'front', "regular 5'",
        align.START_WITHIN_SEQ2,
        align.STOP_WITHIN_SEQ2,
        align.START_WITHIN_SEQ1),
    prefix=AdapterType('prefix', "anchored 5'", align.STOP_WITHIN_SEQ2),
    suffix=AdapterType('suffix', "anchored 3'", align.START_WITHIN_SEQ2),
    anywhere=AdapterType('anywhere', "variable 5'/3'", align.SEMIGLOBAL),
    linked=AdapterType('linked', 'linked', 'linked')
)

def where_int_to_dict(where):
    """Convert a "where" flags integer to a dictionary of values from the
    corresponding AdapterType.
    """
    for adapter_type in ADAPTER_TYPES.values():
        if where == adapter_type.flags:
            return adapter_type.asdict()
    raise ValueError("Invalid WHERE value: {}".format(where))

# TODO get rid of these constants
BACK = ADAPTER_TYPES['back'].flags
FRONT = ADAPTER_TYPES['front'].flags
PREFIX = ADAPTER_TYPES['prefix'].flags
SUFFIX = ADAPTER_TYPES['suffix'].flags
ANYWHERE = ADAPTER_TYPES['anywhere'].flags
LINKED = ADAPTER_TYPES['linked'].flags

# TODO: specify this externally rather than hard-coding
DEFAULT_ADAPTERS_URL = "https://raw.githubusercontent.com/jdidion/atropos/master/atropos/adapters/sequencing_adapters.fa"
DEFAULT_ADAPTERS_PATH = os.path.join(
    os.path.dirname(__file__), 'sequencing_adapters.fa')

class AdapterParser(object):
    """Factory for Adapter classes that all use the same parameters (error rate,
    indels etc.). The given **kwargs will be passed to the Adapter constructors.
    
    Args:
        colorspace: Whether the adapter sequences are in colorspace.
        cache: The AdapterCache to use for fetching/storing named adapters.
    """
    def __init__(self, colorspace=False, cache=None, **kwargs):
        self.colorspace = colorspace
        self.cache = cache
        self.constructor_args = kwargs
        self.adapter_class = ColorspaceAdapter if colorspace else Adapter
    
    def parse(self, spec=None, name=None, cmdline_type='back'):
        """Parse an adapter specification not using ``file:`` notation and return
        an object of an appropriate Adapter class. The notation for anchored
        5' and 3' adapters is supported. If the name parameter is None, then
        an attempt is made to extract the name from the specification
        (If spec is 'name=ADAPTER', name will be 'name'.)
        
        Args:
            spec: The adapter spec.
            name: The adapter name. If not provided, one is automatically
                generated.
            cmdline_type: describes which commandline parameter was used (``-a``
                is 'back', ``-b`` is 'anywhere', and ``-g`` is 'front').
        
        TODO: describe the adapter spec format
        
        Returns:
            An :class:`Adapter` instance.
        """
        if cmdline_type not in ADAPTER_TYPES:
            raise ValueError(
                'cmdline_type cannot be {0!r}'.format(cmdline_type))
        where = ADAPTER_TYPES[cmdline_type].flags
        
        if name is None and spec is None:
            raise ValueError('Either name or spec must be given')
        elif name is None:
            if self.cache and self.cache.has_name(spec):
                name = spec
                spec = self.cache.get_for_name(name)
        elif spec is None:
            if self.cache and self.cache.has_name(name):
                spec = self.cache.get_for_name(name)
        
        if spec is None:
            raise ValueError('Name not found: {}'.format(name))
        elif name is None:
            name, spec = _extract_name_from_spec(spec)
        
        if self.cache and name is not None:
            self.cache.add(name, spec)
        
        sequence = spec
        if where == FRONT and spec.startswith('^'):  # -g ^ADAPTER
            sequence, where = spec[1:], PREFIX
        elif where == BACK:
            sequence1, middle, sequence2 = spec.partition('...')
            if middle == '...':
                if not sequence1:  # -a ...ADAPTER
                    sequence = sequence1[3:]
                elif not sequence2:  # -a ADAPTER...
                    sequence, where = spec[:-3], PREFIX
                else:  # -a ADAPTER1...ADAPTER2
                    if self.colorspace:
                        raise NotImplementedError(
                            'Using linked adapters in colorspace is not '
                            'supported.')
                    if sequence1.startswith('^') or sequence2.endswith('$'):
                        raise NotImplementedError(
                            'Using "$" or "^" when  specifying a linked '
                            'adapter is not supported.')
                    return LinkedAdapter(
                        sequence1, sequence2, name=name,
                        **self.constructor_args)
            elif spec.endswith('$'):   # -a ADAPTER$
                sequence, where = spec[:-1], SUFFIX
        
        if not sequence:
            raise ValueError('The adapter sequence is empty')
        
        return self.adapter_class(
            sequence, where, name=name, **self.constructor_args)
    
    def parse_multi(self, back=None, anywhere=None, front=None):
        """Parse all three types of commandline options that can be used to
        specify adapters. back, anywhere and front are lists of strings,
        corresponding to the respective commandline types (-a, -b, -g).
        
        Args:
            back: Back-adapter specs.
            anywhere: Anywhere-adapter specs.
            front: Front-adapter specs.
        
        Return:
            A list of appropriate Adapter classes.
        """
        adapters = []
        for specs, cmdline_type in (
                (back, 'back'), (anywhere, 'anywhere'), (front, 'front')):
            if not specs:
                continue
            for spec in specs:
                if spec.startswith('file:'):
                    # read adapter sequences from a file
                    with FastaReader(spec[5:]) as fasta:
                        for record in fasta:
                            name = record.name.split(None, 1)[0]
                            adapters.append(
                                self.parse(record.sequence, name, cmdline_type))
                else:
                    adapters.append(
                        self.parse(spec=spec, cmdline_type=cmdline_type))
        return adapters

class Adapter(object):
    """An adapter knows how to match itself to a read. In particular, it knows
    where it should be within the read and how to interpret wildcard characters.
    
    Args:
        sequence: The adapter sequence as string. Will be converted to
            uppercase. Also, Us will be converted to Ts.
        where: One of the BACK, FRONT, PREFIX, SUFFIX or ANYWHERE constants.
            This influences where the adapter is allowed to appear within in the
            read and also which part of the read is removed.
        max_error_rate: Maximum allowed error rate. The error rate is the number
            of errors in the alignment divided by the length of the part of the
            alignment that matches the adapter.
        min_overlap: Minimum length of the part of the alignment that
            matches the adapter.
        read_wildcards: Whether IUPAC wildcards in the read are allowed.
        adapter_wildcards: Whether IUPAC wildcards in the adapter are allowed.
        name: optional name of the adapter. If not provided, the name is set to
            a unique number.
        indels: Whether indels are allowed in adapter matches.
        indel_cost: Cost of indel bases in alignment scoring.
        match_probability: Callabale that computes the probability that a match
            could happen by random chance.
        max_rmp: Maximum random-match probability below which a match is
            considered genuine.
        gc_content: Expected GC content of sequences.
    """
    def __init__(
            self, sequence, where, max_error_rate=0.1, min_overlap=3,
            read_wildcards=False, adapter_wildcards=True, name=None,
            indels=True, indel_cost=1, match_probability=None, max_rmp=None,
            gc_content=0.5):
        self.debug = False
        self.name = _generate_adapter_name() if name is None else name
        self.sequence = parse_braces(sequence.upper().replace('U', 'T'))
        assert len(self.sequence) > 0
        self.where = where
        self.max_error_rate = max_error_rate
        self.min_overlap = min(min_overlap, len(self.sequence))
        self.match_probability = match_probability
        self.max_rmp = max_rmp
        self.gc_content = gc_content
        self.indels = indels
        seq_set = set(self.sequence)
        self.adapter_wildcards = (
            adapter_wildcards and not seq_set <= set('ACGT'))
        if self.adapter_wildcards and not seq_set <= IUPAC_BASES:
            raise ValueError(
                'Invalid character(s) in adapter sequence: {}'.format(
                    ','.join(seq_set - IUPAC_BASES)))
        self.read_wildcards = read_wildcards
        # redirect trimmed() to appropriate function depending on adapter type
        trimmers = {
            FRONT: self._trimmed_front,
            PREFIX: self._trimmed_front,
            BACK: self._trimmed_back,
            SUFFIX: self._trimmed_back,
            ANYWHERE: self._trimmed_anywhere
        }
        self.trimmed = trimmers[where]
        if where == ANYWHERE:
            self._front_flag = None  # means: guess
        else:
            self._front_flag = where not in (BACK, SUFFIX)
        # statistics about length of removed sequences
        self.lengths_front = CountingDict()
        self.lengths_back = CountingDict()
        self.errors_front = NestedDict()
        self.errors_back = NestedDict()
        self.adjacent_bases = { 'A': 0, 'C': 0, 'G': 0, 'T': 0, '': 0 }
        self.aligner = align.Aligner(
            self.sequence, self.max_error_rate, flags=self.where,
            wildcard_ref=self.adapter_wildcards,
            wildcard_query=self.read_wildcards)
        self.aligner.min_overlap = self.min_overlap
        if self.indels:
            self.aligner.indel_cost = indel_cost
        else:
            # TODO
            # When indels are disallowed, an entirely different algorithm
            # should be used.
            self.aligner.indel_cost = 100000
    
    def __repr__(self):
        return '<Adapter(name="{name}", sequence="{sequence}", where={where}, '\
            'max_error_rate={max_error_rate}, min_overlap={min_overlap}, '\
            'read_wildcards={read_wildcards}, '\
            'adapter_wildcards={adapter_wildcards}, '\
            'indels={indels})>'.format(**vars(self))
    
    def enable_debug(self):
        """Print out the dynamic programming matrix after matching a read to an
        adapter.
        """
        self.debug = True
        self.aligner.enable_debug()
    
    def match_to(self, read):
        """Attempt to match this adapter to the given read.
        
        Args:
            read: A :class:`Sequence` instance.
        
        Returns:
            A :class:`Match` instance if a match was found; return None if no
            match was found given the matching criteria (minimum overlap length,
            maximum error rate).
        """
        read_seq = read.sequence.upper()
        
        # try to find an exact match first unless wildcards are allowed
        pos = -1
        if not self.adapter_wildcards:
            if self.where == PREFIX:
                if read_seq.startswith(self.sequence):
                    pos = 0
            elif self.where == SUFFIX:
                if read_seq.endswith(self.sequence):
                    pos = (len(read_seq) - len(self.sequence))
            else:
                pos = read_seq.find(self.sequence)
        
        if pos >= 0:
            seqlen = len(self.sequence)
            return Match(
                0, seqlen, pos, pos + seqlen, seqlen, 0, self._front_flag,
                self, read)
        
        # try approximate matching
        alignment = None
        if not self.indels and self.where in (PREFIX, SUFFIX):
            if self.where == PREFIX:
                alignment = align.compare_prefixes(
                    self.sequence, read_seq,
                    wildcard_ref=self.adapter_wildcards,
                    wildcard_query=self.read_wildcards)
            else:
                alignment = align.compare_suffixes(
                    self.sequence, read_seq,
                    wildcard_ref=self.adapter_wildcards,
                    wildcard_query=self.read_wildcards)
        else:
            alignment = self.aligner.locate(read_seq)
            if self.debug:
                print(self.aligner.dpmatrix)  # pragma: no cover
        
        if alignment:
            astart, astop, rstart, rstop, matches, errors = alignment
            size = astop - astart
            if ((
                    size >=
                    self.min_overlap and errors / size <=
                    self.max_error_rate
                ) and (
                    self.max_rmp is None or
                    self.match_probability(matches, size) <= self.max_rmp)):
                return Match(
                    astart, astop, rstart, rstop, matches, errors,
                    self._front_flag, self, read)
        
        return None
    
    def _trimmed_anywhere(self, match):
        """Trims an adapter from either the front or back of sequence.
        
        Returns:
            A :class:`Sequence` instance: the trimmed read.
        """
        if match.front:
            return self._trimmed_front(match)
        else:
            return self._trimmed_back(match)
    
    def _trimmed_front(self, match):
        """Trims an adapter from the front of sequence.
        
        Returns:
            A :class:`Sequence` instance: the trimmed read.
        """
        # TODO move away
        self.lengths_front[match.rstop] += 1
        self.errors_front[match.rstop][match.errors] += 1
        return match.read[match.rstop:]
    
    def _trimmed_back(self, match):
        """Trims an adapter from the back of sequence.
        
        Returns:
            A :class:`Sequence` instance: the trimmed read.
        """
        # TODO move away
        self.lengths_back[len(match.read) - match.rstart] += 1
        self.errors_back[len(match.read) - match.rstart][match.errors] += 1
        adjacent_base = match.read.sequence[match.rstart-1:match.rstart]
        if adjacent_base not in 'ACGT':
            adjacent_base = ''
        self.adjacent_bases[adjacent_base] += 1
        return match.read[:match.rstart]
    
    def __len__(self):
        return len(self.sequence)
    
    def random_match_probabilities(self):
        """Estimate probabilities that this adapter matches a random sequence. 
        Indels are not taken into account.

        Returns:
            A list of probabilities the same length as this adapter's sequence,
            where the value at position 'i' is the probability that i bases of
            this adapter match a random sequence.
        """
        if self._front_flag:
            seq = self.sequence[::-1]
        else:
            seq = self.sequence
        
        #matches = 0
        base_probs = (self.gc_content, (1 - self.gc_content) / 2.0)
        probabilities = [1.0] + ([0] * len(seq))
        c_bases = frozenset(GC_BASES if self.adapter_wildcards else 'GC')
        
        # TODO: this doesn't work; need to figure out if RandomMatchProbability
        # can be used for this.
        #for idx, base in enumerate(seq, 1):
        #    if base in c_bases:
        #        matches += 1
        #    probabilities[idx] = self.match_probability(
        #        matches, idx, *base_probs)
        
        cur_p = 1.0
        for idx, base in enumerate(seq, 1):
            cur_p *= base_probs[0 if base in c_bases else 1]
            probabilities[idx] = cur_p

        return probabilities
    
    def summarize(self):
        """Summarize the activities of this :class:`Adapter`.
        """
        total_front = sum(self.lengths_front.values())
        total_back = sum(self.lengths_back.values())
        
        stats = MergingDict(
            adapter_class=self.__class__.__name__,
            total_front=total_front,
            total_back=total_back,
            total=total_front + total_back,
            match_probabilities=self.random_match_probabilities())
        
        where = self.where
        assert (
            where in (ANYWHERE, LINKED) or
            (where in (BACK, SUFFIX) and total_front == 0) or
            (where in (FRONT, PREFIX) and total_back == 0))
        
        stats["where"] = where_int_to_dict(where)
        stats["sequence"] = Const(self.sequence)
        stats["max_error_rate"] = Const(self.max_error_rate)
        if where in (ANYWHERE, FRONT, PREFIX):
            stats["lengths_front"] = self.lengths_front
            stats["errors_front"] = self.errors_front
        if where in (ANYWHERE, BACK, SUFFIX):
            stats["lengths_back"] = self.lengths_back
            stats["errors_back"] = self.errors_back
        if where in (BACK, SUFFIX):
            stats["adjacent_bases"] = self.adjacent_bases
        
        return stats

class ColorspaceAdapter(Adapter):
    """An adapter for a colorspace sequence.
    
    Args:
        args, kwargs: Arguments to pass to :class:`Adapter` constructor.
    """
    def __init__(self, *args, **kwargs):
        if kwargs.get('adapter_wildcards', False):
            raise ValueError("Wildcards not supported for colorspace adapters")
        kwargs['adapter_wildcards'] = False
        super(ColorspaceAdapter, self).__init__(*args, **kwargs)
        has_nucleotide_seq = False
        if set(self.sequence) <= set('ACGT'):
            # adapter was given in basespace
            self.nucleotide_sequence = self.sequence
            has_nucleotide_seq = True
            self.sequence = cs.encode(self.sequence)[1:]
        if self.where in (PREFIX, FRONT) and not has_nucleotide_seq:
            raise ValueError(
                "A 5' colorspace adapter needs to be given in nucleotide space")
        self.aligner.reference = self.sequence
    
    def match_to(self, read):
        """Attempt to match this adapter to the given read.
        
        Args:
            read: A :class:`Sequence` instance.
        
        Returns:
            A :class:`Match` instance if a match was found; return None if no
            match was found given the matching criteria (minimum overlap length,
            maximum error rate).
        """
        if self.where != PREFIX:
            return super(ColorspaceAdapter, self).match_to(read)
        # create artificial adapter that includes a first color that encodes the
        # transition from primer base into adapter
        asequence = (
            cs.ENCODE[read.primer + self.nucleotide_sequence[0:1]] +
            self.sequence)
        
        pos = 0 if read.sequence.startswith(asequence) else -1
        if pos >= 0:
            match = Match(
                0, len(asequence), pos, pos + len(asequence),
                len(asequence), 0, self._front_flag, self, read)
        else:
            # try approximate matching
            self.aligner.reference = asequence
            alignment = self.aligner.locate(read.sequence)
            if self.debug:
                print(self.aligner.dpmatrix)  # pragma: no cover
            if alignment is not None:
                match = Match(*(alignment + (self._front_flag, self, read)))
            else:
                match = None
        
        if match is None:
            return None
        assert (
            match.length > 0 and
            match.errors / match.length <= self.max_error_rate)
        assert match.length >= self.min_overlap
        return match

    def _trimmed_front(self, match):
        """Trims an adapter from the front of sequence.
        
        Returns:
            A :class:`Sequence` instance: the trimmed read.
        """
        read = match.read
        self.lengths_front[match.rstop] += 1
        self.errors_front[match.rstop][match.errors] += 1
        # to remove a front adapter, we need to re-encode the first color
        # following the adapter match
        color_after_adapter = read.sequence[match.rstop:match.rstop + 1]
        if not color_after_adapter:
            # the read is empty
            return read[match.rstop:]
        base_after_adapter = cs.DECODE[
            self.nucleotide_sequence[-1:] + color_after_adapter]
        new_first_color = cs.ENCODE[read.primer + base_after_adapter]
        new_read = read[:]
        new_read.sequence = new_first_color + read.sequence[(match.rstop + 1):]
        new_read.qualities = None
        if read.qualities:
            new_read.qualities = read.qualities[match.rstop:]
        return new_read

    def _trimmed_back(self, match):
        """Trims an adapter from the back of sequence.
        
        Returns:
            A :class:`Sequence` instance: the trimmed read.
        """
        # trim one more color if long enough
        adjusted_rstart = max(match.rstart - 1, 0)
        self.lengths_back[len(match.read) - adjusted_rstart] += 1
        self.errors_back[len(match.read) - adjusted_rstart][match.errors] += 1
        return match.read[:adjusted_rstart]

    def __repr__(self):
        return '<ColorspaceAdapter(sequence={0!r}, where={1})>'.format(
            self.sequence, self.where)

# TODO Consolidate Match and a LinkedMatch class.

class LinkedMatch(object):
    """Represent a match of a LinkedAdapter.
    
    Args:
        front_match: The match to the front of the sequence.
        back_match: The match to the back of the sequence.
        adapter: The matched adapter.
    """
    def __init__(self, front_match, back_match, adapter):
        self.front_match = front_match
        self.back_match = back_match
        self.adapter = adapter
        assert front_match is not None
    
    def get_info_record(self):
        """Returns the info record for the either the back or forward match.
        """
        if self.back_match:
            return self.back_match.get_info_record()
        else:
            return self.front_match.get_info_record()

class LinkedAdapter(object):
    """An adapter with linked front and back sequences.
    
    Args:
        front_sequence: Front adapter sequence.
        back_sequence: Back adapter sequence.
        front_anchored: Whether the front adapter is anchored.
        back_anchored: Whether the back adapter is anchored.
        name: Adapter name.
        kwargs: passed on to individual :class:`Adapter` constructors.
    """
    def __init__(
            self, front_sequence, back_sequence, front_anchored=True,
            back_anchored=False, name=None, **kwargs):
        assert front_anchored and not back_anchored
        where1 = PREFIX if front_anchored else FRONT
        where2 = SUFFIX if back_anchored else BACK
        self.front_anchored = front_anchored
        self.back_anchored = back_anchored
        
        # The following attributes are needed for the report
        self.where = LINKED
        self.name = _generate_adapter_name() if name is None else name
        self.front_adapter = Adapter(
            front_sequence, where=where1, name=None, **kwargs)
        self.back_adapter = Adapter(
            back_sequence, where=where2, name=None, **kwargs)
    
    def enable_debug(self):
        """Enable debug on adapters.
        """
        self.front_adapter.enable_debug()
        self.back_adapter.enable_debug()
    
    def match_to(self, read):
        """Match the linked adapters against the given read. If the 'front'
        adapter is not found, the 'back' adapter is not searched for.
        
        Args:
            read: A :class:`Sequence` instance.
        
        Returns:
            A :class:`Match` instance if a match was found; return None if no
            match was found given the matching criteria (minimum overlap length,
            maximum error rate).
        """
        front_match = self.front_adapter.match_to(read)
        if front_match is None:
            return None
        # TODO use match.trimmed() instead as soon as that does not update
        # statistics anymore
        read = read[front_match.rstop:]
        back_match = self.back_adapter.match_to(read)
        return LinkedMatch(front_match, back_match, self)

    def trimmed(self, match):
        """Returns the read trimmed with the front and/or back adapter
        trimmer(s).
        
        Args:
            match: The match to trim.
        
        Returns:
            The trimmed read.
        """
        front_trimmed = self.front_adapter.trimmed(match.front_match)
        if match.back_match:
            return self.back_adapter.trimmed(match.back_match)
        else:
            return front_trimmed
    
    def summarize(self):
        """Returns the summary dict for this adapter.
        """
        total_front = sum(self.front_adapter.lengths_front.values())
        total_back = sum(self.back_adapter.lengths_back.values())
        
        stats = MergingDict(
            total_front=total_front,
            total_back=total_back,
            total=total_front + total_back)
        
        where = self.where
        assert (
            where in (ANYWHERE, LINKED) or
            (where in (BACK, SUFFIX) and total_front == 0) or
            (where in (FRONT, PREFIX) and total_back == 0))
        
        stats["where"] = where_int_to_dict(where)
        stats["front_sequence"] = Const(self.front_adapter.sequence)
        stats["front_match_probabilities"] = Const(
            self.front_adapter.random_match_probabilities())
        stats["back_sequence"] = Const(self.back_adapter.sequence)
        stats["back_match_probabilities"] = Const(
            self.back_adapter.random_match_probabilities())
        stats["front_max_error_rate"] = Const(self.front_adapter.max_error_rate)
        stats["back_max_error_rate"] = Const(self.back_adapter.max_error_rate)
        stats["front_lengths_front"] = self.front_adapter.lengths_front
        stats["front_lengths_back"] = self.front_adapter.lengths_back
        stats["back_lengths_front"] = self.back_adapter.lengths_front
        stats["back_lengths_back"] = self.back_adapter.lengths_back
        # have to clone these nested dicts and set them
        # up with a custom merge function
        stats["front_errors_front"] = self.front_adapter.errors_front
        stats["front_errors_back"] = self.front_adapter.errors_back
        stats["back_errors_front"] = self.back_adapter.errors_front
        stats["back_errors_back"] = self.back_adapter.errors_back
        
        return stats

class AdapterCache(object):
    """Cache for known adapters.
    
    Args:
        path: Path to file where cache is written.
        auto_reverse_complement: Whether adapter reverse-complements should
            automatically be added to the cache.
    """
    def __init__(self, path=".adapters", auto_reverse_complement=False):
        self.path = path
        self.auto_reverse_complement = auto_reverse_complement
        if path and os.path.exists(path):
            with open(path, "rb") as cache:
                self.seq_to_name, self.name_to_seq = pickle.load(cache)
        else:
            self.seq_to_name = {}
            self.name_to_seq = {}
    
    @property
    def empty(self):
        """Whether the cache is empty.
        """
        return len(self.seq_to_name) == 0
    
    def save(self):
        """Save the cache to file.
        """
        if self.path is not None:
            with open(self.path, "wb") as cache:
                pickle.dump((self.seq_to_name, self.name_to_seq), cache)
    
    def add(self, name, seq):
        """Add a sequence to the cache.
        
        Args:
            name: Adapter name.
            seq: Adapter sequence.
        """
        self._add(name, seq)
        if self.auto_reverse_complement:
            self._add("{}_rc".format(name), reverse_complement(seq))
    
    def _add(self, name, seq):
        if seq not in self.seq_to_name:
            self.seq_to_name[seq] = set()
        self.seq_to_name[seq].add(name)
        self.name_to_seq[name] = seq
    
    def load_from_file(self, path=DEFAULT_ADAPTERS_PATH):
        """Load cached data from a file.
        
        Args:
            path: Path from which to load.
        """
        with open(path, "rt") as i:
            return self.load_from_fasta(i)

    def load_from_url(self, url=DEFAULT_ADAPTERS_URL):
        """Load adapter data from a URL.
        
        Args:
            url: URL from which to load.
        """
        logging.getLogger().info(
            "Loading list of known contaminants from %s", url)
        try:
            fasta = urlopen(url).read().decode().split("\n")
            return self.load_from_fasta(fasta)
        except URLError:
            if url.startswith("file:"):
                url = url[5:]
            return self.load_from_file(url)
    
    def load_from_fasta(self, fasta):
        """Load adapter data from a FASTA file.
        
        Args:
            fasta: FASTA file.
        """
        close = False
        if isinstance(fasta, str):
            fasta = open(fasta, 'rt')
            close = True
        num_records = None
        with FastaReader(fasta) as fasta:
            for num_records, record in enumerate(fasta, 1):
                name = record.name.split(None, 1)[0]
                seq = record.sequence
                self.add(name, seq)
        if close:
            fasta.close()
        return num_records
    
    def load_default(self):
        """Tries to load from default URL first, then from default path.
        """
        try:
            return self.load_from_url()
        except (OSError, IOError):
            logging.getLogger().warning(
                "Error loading adapters from URL %s; loading from file",
                DEFAULT_ADAPTERS_URL)
        try:
            return self.load_from_file()
        except IOError:
            logging.getLogger().warning(
                "Error loading adapters from file %s; loading from file",
                DEFAULT_ADAPTERS_PATH)
    
    @property
    def names(self):
        """Sequence of adapter names.
        """
        return list(self.name_to_seq.keys())
    
    @property
    def sequences(self):
        """Sequence of adapter sequences.
        """
        return list(self.seq_to_name.keys())
    
    def iter_names(self):
        """Returns an iterator over adapter names.
        """
        return self.name_to_seq.items()
    
    def iter_sequences(self):
        """Returns an iterator over adapter sequences.
        """
        return self.seq_to_name.items()
    
    def has_name(self, name):
        """Returns whether this cache contains the specified name.
        
        Args:
            name: The adapter name.
        """
        return name in self.name_to_seq
    
    def get_for_name(self, name):
        """Returns the sequence associated with a name.
        
        Args:
            name: The name to fetch.
        
        Returns:
            The sequence.
        """
        return self.name_to_seq[name]
    
    def has_seq(self, seq):
        """Tests whether a sequence is in the cache.
        
        Args:
            seq: The sequence to check.
        
        Returns:
            True if the sequence is in the cache.
        """
        return seq in self.seq_to_name
    
    def get_for_seq(self, seq):
        """Returns the name associated with a given sequence.
        
        Args:
            seq: The sequence to fetch.
        
        Returns:
            The name associated with the sequence.
        """
        return list(self.seq_to_name[seq])

def parse_braces(sequence):
    """Replace all occurrences of ``x{n}`` (where x is any character) with n
    occurrences of x. Raise ValueError if the expression cannot be parsed.
    
    Examples:
        >>> parse_braces('TGA{5}CT')
        TGAAAAACT
    """
    # Simple DFA with four states, encoded in prev
    result = ''
    prev = None
    for char in re.split(r'(\{|\})', sequence):
        if char == '':
            continue
        if prev is None:
            if char == '{':
                raise ValueError('"{" must be used after a character')
            if char == '}':
                raise ValueError('"}" cannot be used here')
            prev = char
            result += char
        elif prev == '{':
            prev = int(char)
            if not 0 <= prev <= 10000:
                raise ValueError('Value {} invalid'.format(prev))
        elif isinstance(prev, int):
            if char != '}':
                raise ValueError('"}" expected')
            result = result[:-1] + result[-1] * prev
            prev = None
        else:
            if char != '{':
                raise ValueError('Expected "{"')
            prev = '{'
    # Check if we are in a non-terminating state
    if isinstance(prev, int) or prev == '{':
        raise ValueError("Unterminated expression")
    return result

def _extract_name_from_spec(spec):
    """Parse an adapter specification given as 'name=adapt' into 'name'
    and 'adapt'.
    
    Args:
        spec: Adapter spec.
    
    Returns:
        (name, spec)
    """
    fields = spec.split('=', 1)
    if len(fields) > 1:
        name, spec = fields
        name = name.strip()
    else:
        name = None
    spec = spec.strip()
    return name, spec

ADAPTER_ID_GENERATOR = itertools.count(1)

def _generate_adapter_name():
    return str(next(ADAPTER_ID_GENERATOR))
